import importlib
import argparse
import logging

logging.basicConfig(format='%(message)s')
logging.getLogger().setLevel(logging.INFO)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Initializes environment for distributed LightGBM')
    parser.add_argument('config_file', help='LightGBM config file')
    parser.add_argument('mlist_file', help='LightGBM machine list file')
    args = parser.parse_args()
    utils = importlib.import_module("utils", package=".")
    print(utils.init_lightgbm_env(args.config_file, args.mlist_file))



import docker
from docker.errors import DockerException
from kubeflow.fairing.functions.function_shim import get_execution_obj_type, ObjectType
from kubeflow.fairing.preprocessors.function import FunctionPreProcessor
from kubeflow.fairing.preprocessors.base import BasePreProcessor
from kubeflow.fairing.preprocessors.full_notebook import FullNotebookPreProcessor

def guess_preprocessor(entry_point, input_files, output_map):
    """Preprocessor to use to modify inputs before sending them to docker build

    :param entry_point: entry_point which to use
    :param input_files: input files
    :param output_map: output

    """
    if get_execution_obj_type(entry_point) != ObjectType.NOT_SUPPORTED:
        return FunctionPreProcessor(function_obj=entry_point,
                                    input_files=input_files,
                                    output_map=output_map)
    elif isinstance(entry_point, str) and entry_point.endswith(".py"):
        input_files.append(entry_point)
        return BasePreProcessor(executable=entry_point,
                                input_files=input_files,
                                output_map=output_map)
    elif isinstance(entry_point, str) and entry_point.endswith(".ipynb"):
        return FullNotebookPreProcessor(notebook_file=entry_point, input_files=input_files
                                        , output_map=output_map)
    else:
        # TODO (karthikv2k) Handle other entrypoints like python files and full notebooks
        raise NotImplementedError(
            "obj param should be a function or a class, got {}".format(type(entry_point)))


def is_docker_daemon_exists():
    """To check if docker daemon exists or not. """
    try:
        docker.APIClient(version='auto')
        return True
    except DockerException:
        return False



from timeit import default_timer as timer
import httplib2
import os
import logging

from containerregistry.client import docker_creds
from containerregistry.client import docker_name
from containerregistry.client.v2_2 import append
from containerregistry.client.v2_2 import docker_image as v2_2_image
from containerregistry.client.v2_2 import docker_session
from containerregistry.transport import transport_pool
from containerregistry.transform.v2_2 import metadata

from kubeflow.fairing.builders.base_builder import BaseBuilder
from kubeflow.fairing.constants import constants

logger = logging.getLogger(__name__)


class AppendBuilder(BaseBuilder):
    """Builds a docker image by appending a new layer tarball to an existing
    base image. Does not require docker and runs in userspace.

    :param base_image: Base image to use for the build (default: {constants.DEFAULT_BASE_IMAGE})
    :param image_name: image name to use for the new image(default: {constants.DEFAULT_IMAGE_NAME})
    :param preprocessor: Preprocessor{BasePreProcessor} to use to modify inputs
        before sending them to docker build
    :param push: Whether or not to push the image to the registry

    """

    def __init__(self,
                 registry=None,
                 image_name=constants.DEFAULT_IMAGE_NAME,
                 base_image=constants.DEFAULT_BASE_IMAGE,
                 push=True,
                 preprocessor=None):
        super().__init__(
            registry=registry,
            image_name=image_name,
            base_image=base_image,
            push=push,
            preprocessor=preprocessor,
        )

    def build(self):
        """Will be called when the build needs to start"""
        transport = transport_pool.Http(httplib2.Http)
        src = docker_name.Tag(self.base_image, strict=False)
        logger.warning("Building image using Append builder...")
        start = timer()
        new_img = self._build(transport, src)
        end = timer()
        logger.warning("Image successfully built in {}s.".format(end-start))
        dst = docker_name.Tag(
            self.full_image_name(self.context_hash), strict=False)
        if self.push:
            self.timed_push(transport, src, new_img, dst)
        else:
            # TODO(r2d4):
            # Load image into local daemon. This wouldn't provide any speedup
            # over using the docker daemon directly.
            pass

    def _build(self, transport, src):
        file, hash = self.preprocessor.context_tar_gz()  # pylint:disable=redefined-builtin
        self.context_file, self.context_hash = file, hash
        self.image_tag = self.full_image_name(self.context_hash)
        creds = docker_creds.DefaultKeychain.Resolve(src)
        with v2_2_image.FromRegistry(src, creds, transport) as src_image:
            with open(self.context_file, 'rb') as f:
                new_img = append.Layer(src_image, f.read(), overrides=metadata.Overrides(
                    cmd=self.preprocessor.get_command(),
                    user='0', env={"FAIRING_RUNTIME": "1"}))
        return new_img

    def _push(self, transport, src, img, dst):
        creds = docker_creds.DefaultKeychain.Resolve(dst)
        with docker_session.Push(dst, creds, transport,
                                 mount=[src.as_repository()]) as session:
            logger.warning("Uploading {}".format(self.image_tag))
            session.upload(img)
        os.remove(self.context_file)

    def timed_push(self, transport, src, img, dst):
        """Push image to the registry and log the time spent to the log

        :param transport: the http transport to use for sending requests
        :param src: repo from which to mount blobs
        :param img: the image to be pushed
        :param dst: the fully-qualified name of the tag to push

        """
        logger.warning("Pushing image {}...".format(self.image_tag))
        start = timer()
        self._push(transport, src, img, dst)
        end = timer()
        logger.warning(
            "Pushed image {} in {}s.".format(self.image_tag, end-start))



import six
import abc
from google.cloud import storage
from urllib.parse import urlparse

def lookup_storage_class(url):
    scheme = urlparse(url).scheme
    if scheme in ["gcs", "gs"]:
        return GCSStorage
    else:
        return None

def get_storage_class(url):
    res = lookup_storage_class(url)
    if res:
        return res
    else:
        raise RuntimeError(
            "can't find a suitable storage class for {}".format(url))


@six.add_metaclass(abc.ABCMeta)
class Storage:

    # Using shell commands to do the file copy instead of using python libs
    # CLIs like gsutil, s3cmd are optimized and can be easily configured by
    # the user using boto.cfg in the base docker image.
    @abc.abstractmethod
    def copy_cmd(self, src_url, dst_url, recursive=True):
        """gets a command to copy files from/to remote storage from/to local FS"""
        raise NotImplementedError('Storage.copy_cmd')

    @abc.abstractmethod
    def exists(self, url):
        """checks if the url exists in the given storage"""
        raise NotImplementedError('Storage.exists')


class GCSStorage(Storage):

    def __init__(self):
        self.client = storage.Client()

    def copy_cmd(self, src_url, dst_url, recursive=True):
        if recursive:
            rcmd = "-r"
        else:
            rcmd = ""
        return "gsutil cp {} {} {}".format(rcmd, src_url, dst_url)

    @classmethod
    def _check_prefix(cls, bucket, prefix):
        # url points to a dir like resource
        # Checking if at least one blob exists
        blobs = [x for x in bucket.list_blobs(prefix=prefix, max_results=1)]
        if len(blobs) == 1:
            return True
        return False

    def exists(self, url):
        url_parts = urlparse(url)
        bucket_name = url_parts.netloc
        bucket = self.client.bucket(bucket_name)
        if not url_parts.path or url_parts.path == "/":
            return True
        if bucket:
            prefix_key = url_parts.path[1:]
            if prefix_key.endswith("/"):
                return GCSStorage._check_prefix(bucket, prefix_key)
            elif bucket.get_blob(prefix_key):
                return True
            else:
                return GCSStorage._check_prefix(bucket, prefix_key + "/")
        return False



import os
from kubeflow.fairing.ml_tasks.tasks import TrainJob, PredictionEndpoint

if os.getenv('FAIRING_RUNTIME', None) is not None:
    from kubeflow.fairing.runtime_config import config
else:
    from kubeflow.fairing.config import config

name = "fairing"

__version__ = "0.7.0"



import cloudpickle
import logging
import os
import sys
import tempfile

from kubeflow import fairing
from kubeflow.fairing.constants import constants
from kubeflow.fairing.preprocessors.base import BasePreProcessor
from kubeflow.fairing.functions.function_shim import get_execution_obj_type, ObjectType
from kubeflow.fairing.notebook import notebook_util

logger = logging.getLogger(__name__)

FUNCTION_SHIM = 'function_shim.py'
SERIALIZED_FN_FILE = 'pickled_fn.p'

# TODO(@karthikv2k): Ref #122 Find a better way to support deployer specific preprocessing
OUTPUT_FILE = """import cloudpickle
{OBJ_NAME} = cloudpickle.load(open("{SERIALIZED_FN_FILE}", "rb"))
"""


class FunctionPreProcessor(BasePreProcessor):
    """FunctionPreProcessor preprocesses a single function.
       It sets as the command a function_shim that calls the function directly.
       :param BasePreProcessor: a context that gets sent to the builder for the docker build
       and sets the entrypoint.
    """

    def __init__(self,
                 function_obj,
                 path_prefix=constants.DEFAULT_DEST_PREFIX,
                 output_map=None,
                 input_files=None):
        """Init the function preprocess class
        :param function_obj: the name of the function or class to be called.
        :param path_prefix: the defaut destion path prefix '/app/'.
        :param output_map: a dict of files to be added without preprocessing.
        :param input_files: the source files to be processed.

        """
        super().__init__(
            output_map=output_map,
            path_prefix=path_prefix,
            input_files=input_files)

        if not notebook_util.is_in_notebook():
            logger.warning("The FunctionPreProcessor is optimized for using in a notebook or "
                           "IPython environment. For it to work, the python version should be "
                           "same for both local python and the python in the docker. Please look "
                           "at alternatives like BasePreprocessor or FullNotebookPreprocessor.")

        if get_execution_obj_type(function_obj) == ObjectType.NOT_SUPPORTED:
            raise RuntimeError("Object must of type function or a class")

        fairing_dir = os.path.dirname(fairing.__file__)
        self.output_map[os.path.join(fairing_dir, "functions", FUNCTION_SHIM)] = \
            os.path.join(path_prefix, FUNCTION_SHIM)

        # Make sure fairing can use imported as a module
        self.output_map[os.path.join(fairing_dir, '__init__.py')] = \
            os.path.join(path_prefix, "fairing", '__init__.py')

        # Make sure cloudpickle can be imported as a module
        cloudpickle_dir = os.path.dirname(cloudpickle.__file__)
        self.output_map[os.path.join(cloudpickle_dir, '__init__.py')] = \
            os.path.join(path_prefix, "cloudpickle", '__init__.py')
        self.output_map[os.path.join(cloudpickle_dir, 'cloudpickle.py')] = \
            os.path.join(path_prefix, "cloudpickle", 'cloudpickle.py')

        _, temp_payload_file = tempfile.mkstemp()
        with open(temp_payload_file, "wb") as f:
            cloudpickle.dump(function_obj, f)
        # Adding the serialized file to the context
        payload_file_in_context = os.path.join(path_prefix, SERIALIZED_FN_FILE)
        self.output_map[temp_payload_file] = payload_file_in_context

        # TODO(@karthikv2k): Ref #122 Find a better way to support deployer specific preprocessing
        _, temp_payload_wrapper_file = tempfile.mkstemp()
        with open(temp_payload_wrapper_file, "w") as f:
            contents = OUTPUT_FILE.format(OBJ_NAME=function_obj.__name__,
                                          SERIALIZED_FN_FILE=SERIALIZED_FN_FILE)
            f.write(contents)
        # Adding the serialized file to the context
        payload_wrapper_file_in_context = os.path.join(
            path_prefix, function_obj.__name__ + ".py")
        self.output_map[temp_payload_wrapper_file] = payload_wrapper_file_in_context

        local_python_version = ".".join(
            [str(x) for x in sys.version_info[0:2]])

        self.command = ["python", os.path.join(self.path_prefix, FUNCTION_SHIM),
                        "--serialized_fn_file", payload_file_in_context,
                        "--python_version", local_python_version]

    def get_command(self):
        """ Get the execute python command.
        :returns: command: the command line will be executed
        """
        return self.command



import os

from kubeflow.fairing.preprocessors.base import BasePreProcessor
from kubeflow.fairing.constants import constants
from kubeflow.fairing.notebook import notebook_util


class FullNotebookPreProcessor(BasePreProcessor):
    """ The Full notebook preprocess for the context which comes from BasePreProcessor.
    :param BasePreProcessor: a context that gets sent to the builder for the docker build and
    sets the entrypoint
    """
    # TODO: Allow configuration of errors / timeout options
    def __init__(self,
                 notebook_file=None,
                 output_file="fairing_output_notebook.ipynb",
                 input_files=None,
                 command=None,
                 path_prefix=constants.DEFAULT_DEST_PREFIX,
                 output_map=None):
        """ Init the full notebook preprocess.
        :param notebook_file: the jupyter notebook file.
        :param output_file: the output file, the defaut name is 'fairing_output_notebook.ipynb'.
        :param input_files: the source files to be processed.
        :param command: the command to pass to the builder.
        :param path_prefix: the defaut destion path prefix '/app/'.
        :param output_map: a dict of files to be added without preprocessing.

        """
        if notebook_file is None and notebook_util.is_in_notebook():
            notebook_file = notebook_util.get_notebook_name()

        if notebook_file is None:
            raise ValueError('A notebook_file must be provided.')

        relative_notebook_file = notebook_file
        # Convert absolute notebook path to relative path
        if os.path.isabs(notebook_file[0]):
            relative_notebook_file = os.path.relpath(notebook_file)

        if command is None:
            command = ["papermill", relative_notebook_file, output_file, "--log-output"]

        input_files = input_files or []
        if relative_notebook_file not in input_files:
            input_files.append(relative_notebook_file)

        super().__init__(
            executable=None,
            input_files=input_files,
            command=command,
            output_map=output_map,
            path_prefix=path_prefix)


    def set_default_executable(self):
        """Ingore the default executable setting for the full_notebook preprocessor.
        """
        pass





class RuntimeConfig(object):
    """A passthrough config shim that runs in the fairing runtime"""
    def __init__(self):
        pass

    def set_preprocessor(self, name, **kwargs): #pylint:disable=unused-argument
        self._preprocessor = name

    def get_preprocessor(self):
        return self._preprocessor

    def set_builder(self, name, **kwargs): #pylint:disable=unused-argument
        self._builder = name

    def get_builder(self):
        return self._builder

    def set_deployer(self, name, **kwargs): #pylint:disable=unused-argument
        self._deployer = name

    def get_deployer(self, **kwargs): #pylint:disable=unused-argument
        return self._deployer

    def run(self):
        pass

    def reset(self):
        pass

    def fn(self, func): #pylint:disable=no-self-use
        return func

config = RuntimeConfig()



import os
import zlib
import uuid


def get_image(repository, name):
    """Get the full image name by integrating repository and image name.

    :param repository: The name of repository
    :param name: The short image name
    :returns: str: Full image name, format: repo/name.

    """
    return "{repo}/{name}".format(
        repo=repository,
        name=name
    )

def is_running_in_k8s():
    """Check if running in the kubernetes cluster."""
    return os.path.isdir('/var/run/secrets/kubernetes.io/')

def get_current_k8s_namespace():
    """Get the current namespace of kubernetes."""
    with open('/var/run/secrets/kubernetes.io/serviceaccount/namespace', 'r') as f:
        return f.readline()

def get_default_target_namespace():
    """Get the default target namespace, if running in the kubernetes cluster,
    will be current namespace, Otherwiase, will be "default".
    """
    if not is_running_in_k8s():
        return 'default'
    return get_current_k8s_namespace()

def crc(file_name):
    """Compute a running Cyclic Redundancy Check checksum.

    :param file_name: The file name that's for crc checksum.

    """
    prev = 0
    for eachLine in open(file_name, "rb"):
        prev = zlib.crc32(eachLine, prev)
    return "%X" % (prev & 0xFFFFFFFF)

def random_tag():
    """Get a random tag."""
    return str(uuid.uuid4()).split('-')[0]



from googleapiclient import discovery
from googleapiclient import errors

from kubeflow.fairing import utils
from kubeflow.fairing import http_utils
from kubeflow.fairing.deployers.deployer import DeployerInterface
from kubeflow.fairing.cloud.gcp import guess_project_name

class GCPJob(DeployerInterface):
    """Handle submitting training job to GCP."""

    def __init__(self, project_id=None, region=None, scale_tier=None, job_config=None):
        """
        :param project_id: Google Cloud project ID to use.
        :param region: region in which the job has to be deployed.
            Ref: https://cloud.google.com/compute/docs/regions-zones/
        :param scale_tier: machine type to use for the job.
            Ref: https://cloud.google.com/ml-engine/docs/tensorflow/machine-types
        :param job_config: Custom job configuration options. If an option is specified
            in the job_config and as a top-level parameter, the parameter overrides
            the value in the job_config.
            Ref: https://cloud.google.com/ml-engine/reference/rest/v1/projects.jobs
        """
        self._project_id = project_id or guess_project_name()
        self._region = region or 'us-central1'
        self._job_config = job_config or {}
        self.scale_tier = scale_tier
        self._ml = discovery.build('ml', 'v1')
        self._ml._http = http_utils.configure_http_instance(self._ml._http) #pylint:disable=protected-access

    def create_request_dict(self, pod_template_spec):
        """Return the request to be sent to the ML Engine API.

        :param pod_template_spec: pod spec template of the training job

        """
        # TODO: Update deploy interface to pass image directly instad of
        # PodTemplateSpec.
        # Retrieve image uri from pod template spec.
        image_uri = pod_template_spec.containers[0].image
        self._job_name = 'fairing_job_{}'.format(utils.random_tag())

        # Merge explicitly specified parameters with the job config dictionary
        request_dict = self._job_config
        request_dict['jobId'] = self._job_name
        if 'trainingInput' not in request_dict:
            request_dict['trainingInput'] = {}

        if self.scale_tier:
            request_dict['trainingInput']['scaleTier'] = self.scale_tier

        if 'masterConfig' not in request_dict['trainingInput']:
            request_dict['trainingInput']['masterConfig'] = {}
        request_dict['trainingInput']['masterConfig']['imageUri'] = image_uri

        if self._region:
            request_dict['trainingInput']['region'] = self._region

        return request_dict

    def deploy(self, pod_template_spec):
        """Deploys the training job

        :param pod_template_spec: pod spec template of the training job

        """
        request_dict = self.create_request_dict(pod_template_spec)

        try:
            print('Creating training job with the following options: {}'.format(
                str(request_dict)))
            response = self._ml.projects().jobs().create( #pylint:disable=unused-variable
                parent='projects/{}'.format(self._project_id),
                body=request_dict
            ).execute()
            print('Job submitted successfully.')
            self.get_logs()
            return self._job_name
        except errors.HttpError as err:
            print('There was an error submitting the job.')
            print(err._get_reason()) #pylint:disable=protected-access

    def get_logs(self):
        """Streams the logs for the training job"""
        print('Access job logs at the following URL:')
        print('https://console.cloud.google.com/mlengine/jobs/{}?project={}'
              .format(self._job_name, self._project_id))



import socket
import json
import os
import time
import tempfile
import configparser
import logging

logger = logging.getLogger(__name__)


def parse_cluster_spec_env():
    """ parse cluster spec env variables"""
    env_config = os.environ.get('CLUSTER_SPEC') or os.environ.get('TF_CONFIG')
    cluster = None
    if env_config:
        env_config_json = json.loads(env_config)
        cluster = env_config_json.get('cluster')
    if not cluster:
        return [], []
    hosts = []
    ports = []
    for val in cluster.values():
        if not isinstance(val, list):
            raise ValueError("Expecting list type values in cluster spec dict. \
                             cluster spec: {}".format(env_config))
        for item in val:
            host, port = item.split(":")
            hosts.append(host)
            ports.append(port)
    ips = [nslookup(host) for host in hosts]

    # Finding the RANK of the current node using hostname or ip address of the current node
    cur_hostname = socket.gethostname()
    cur_host_ip = socket.gethostbyname(cur_hostname)
    if cur_hostname and cur_hostname in hosts:
        rank = hosts.index(cur_hostname)
    elif cur_host_ip and cur_host_ip in ips:
        rank = ips.index(cur_host_ip)
    else:
        print("Can't find rank for current host (hostname={}, \
              ip={})".format(cur_hostname, cur_host_ip))
        rank = 0

    return hosts, ips, ports, rank


def nslookup(hostname, retries=600):
    """Does nslookup for the hostname and returns the IPs for it.

    :param hostname:  hostname to be looked up
    :param retries:  Number of retries before failing. In autoscaled cluster,
                    it might take upto 10mins to create a new node so the default value
                    is set high.(Default value = 600)

    """
    last_exception = None
    for _ in range(retries):
        try:
            time.sleep(1)
            ais = socket.getaddrinfo(hostname, 0, 0, 0, 0)
            return ais[0][-1][0]
        except Exception as e:
            logger.info("not able to lookup ip for %s", hostname)
            last_exception = e
    if last_exception:
        raise last_exception #pylint:disable=raising-bad-type


def write_ip_list_file(file_name, ips, port=None):
    """write list of ips into a file

    :param file_name: fine name
    :param ips: list of ips
    :param port:  default port(Default value = None)

    """
    with open(file_name, 'w') as f:
        if port:
            content = ["{} {}".format(ip, port) for ip in ips]
            content.sort()
            f.write("\n".join(content))
        else:
            sorted_ips = ips.copy().sort()
            f.write("\n".join(sorted_ips))
        f.write("\n")


def load_properties_config_file(config_file):
    """load config from a file

    :param config_file: config file path

    """
    config_parser = configparser.ConfigParser(allow_no_value=True)
    with open(config_file, 'r') as f:
        # Hack to get config parser work with properties files
        buf = "[default]\n" + f.read()
        config_parser.read_string(buf)
        config = dict(config_parser.items(section="default"))
    return config


def save_properties_config_file(config, file_name=None):
    """save config into a file

    :param config: dictionary of give configs
    :param file_name:  path to config file(Default value = None)

    """
    if not file_name:
        _, file_name = tempfile.mkstemp()

    with open(file_name, 'w') as fh:
        content = []
        for k, v in config.items():
            if isinstance(v, bool):
                v = str(v).lower()
            content.append("{}={}".format(k, v))
        fh.write("\n".join(content))
        fh.write("\n")
    return file_name


def update_config_file(file_name, field_name, new_value):
    """update config file

    :param file_name: file name
    :param field_name: field name
    :param new_value: new value to be added/updated

    """
    config = load_properties_config_file(file_name)
    config[field_name] = new_value
    save_properties_config_file(config, file_name)


def scrub_fields(config, filed_names):
    """scrub fields in config

    :param config: config spec
    :param filed_names: name of fields

    """
    for field in filed_names:
        if field in config:
            config.pop(field)
            logger.info("Ignoring %s filed in the config", field)


def get_config_value(config, field_names):
    """get value for a config entry

    :param config:
    :param field_names:

    """
    buf = {}
    for field in field_names:
        if field in config:
            buf[field] = config.get(field)
    if len(buf.items()) > 1:
        raise RuntimeError(
            "More than one field alias is specified: {}".format(buf))
    if len(buf.items()) > 0: #pylint:disable=no-else-return,len-as-condition
        return list(buf.items())[0]
    else:
        return None, None


def init_lightgbm_env(config_file, mlist_file):
    """initialize env for lightgbm

    :param config_file: path to config path
    :param mlist_file: path to file to write ip list

    """
    hosts, ips, ports, rank = parse_cluster_spec_env()
    #print("Rank: {}".format(rank))
    if len(set(ports)) > 1:
        raise RuntimeError(
            "Expecting cluster spec to have same port for all instances/pods."
            " Ports: {}".format(ports))
    port = ports[0]
    write_ip_list_file(mlist_file, ips, port)
    update_config_file(config_file, "local_port", port)
    logger.info("Cluster setup:")
    for x, y, z in zip(hosts, ips, ports):
        logger.info("{}\t{}\t{}".format(x, y, z)) #pylint:disable=logging-format-interpolation
    return rank



import logging
import nbconvert
import re
from nbconvert.preprocessors import Preprocessor as NbPreProcessor
from pathlib import Path

from kubeflow.fairing.preprocessors.base import BasePreProcessor
from kubeflow.fairing.notebook import notebook_util
from kubeflow.fairing.constants import constants


class FilterMagicCommands(NbPreProcessor):
    """Notebook preprocessor that have a comment which started with '!' or '%'.
    :param NbPreProcessor: the notebook preprocessor.
    """
    _magic_pattern = re.compile('^!|^%')

    def filter_magic_commands(self, src):
        """Filter out the source commands with magic pattern.

        :param src: the source commands.
        :returns: filtered: the filtered commands list.

        """
        filtered = []
        for line in src.splitlines():
            match = self._magic_pattern.match(line)
            if match is None:
                filtered.append(line)
        return '\n'.join(filtered)

    def preprocess_cell(self, cell, resources, index): #pylint:disable=unused-argument
        """preprocessor that includes cells

        :param: cell: the notebook cell.
        :param: resources: the code source of the notebook cell.
        :param: index: unused argumnet.
        :return: cell,resources: the notebook cell and its filtered with magic pattern commands.

        """
        if cell['cell_type'] == 'code':
            cell['source'] = self.filter_magic_commands(cell['source'])
        return cell, resources


class FilterIncludeCell(NbPreProcessor):
    """Notebook preprocessor that only includes cells that have a comment 'fairing:include-cell'.
    :param NbPreProcessor: the notebook preprocessor.
    """
    _pattern = re.compile('.*fairing:include-cell.*')

    def filter_include_cell(self, src):
        """Filter the cell that have a comment 'fairing:include-cell'.

        :param: src: the source cell.
        :returns: src: if the source cell matched the filter pattern, or Null.

        """
        for line in src.splitlines():
            match = self._pattern.match(line)
            if match:
                return src
        return ''

    def preprocess_cell(self, cell, resources, index): #pylint:disable=unused-argument
        """ Preprocess the notebook cell.

        :param cell: the notebook cell
        :param resources: the code source of the notebook cell.
        :param index: unused argumnet.
        :return: cell,resources: the notebook cell and its filtered with magic pattern commands.

        """
        if cell['cell_type'] == 'code':
            cell['source'] = self.filter_include_cell(cell['source'])

        return cell, resources


class ConvertNotebookPreprocessor(BasePreProcessor):
    """Convert the notebook preprocessor.
    :param BasePreProcessor: a context that gets sent to the builder for the docker build
    and sets the entrypoint.
    """
    def __init__(self, #pylint:disable=dangerous-default-value
                 notebook_file=None,
                 notebook_preprocessor=FilterMagicCommands,
                 executable=None,
                 command=["python"],
                 path_prefix=constants.DEFAULT_DEST_PREFIX,
                 output_map=None,
                 overwrite=True):
        """The init function ConvertNotebookPreprocessor class.
        :param notebook_file: the notebook file.
        :param notebook_preprocessor: the class FilterMagicCommands.
        :param executable: the file to execute using command (e.g. main.py)
        :param command: the python command.
        :param path_prefix: the defaut destion path prefix '/app/'.
        :param output_map: a dict of files to be added without preprocessing.
        """

        super().__init__(
            executable=executable,
            input_files=[],
            output_map=output_map,
            path_prefix=path_prefix)

        if notebook_file is None and notebook_util.is_in_notebook():
            notebook_file = notebook_util.get_notebook_name()

        self.notebook_file = notebook_file
        self.notebook_preprocessor = notebook_preprocessor
        self.overwrite = overwrite

    def preprocess(self):
        """Preprocessor the Notebook
        :return:[]: the converted notebook list.
        """
        exporter = nbconvert.PythonExporter()
        exporter.register_preprocessor(self.notebook_preprocessor, enabled=True)
        contents, _ = exporter.from_filename(self.notebook_file)
        converted_notebook = Path(self.notebook_file).with_suffix('.py')
        if converted_notebook.exists() and not self.overwrite:
            raise Exception('Default path {} exists but overwrite flag\
                            is False'.format(converted_notebook))
        with open(converted_notebook, 'w') as f:
            logging.info('Converting {} to {}'.format(self.notebook_file, converted_notebook))
            f.write(contents)
        self.executable = converted_notebook
        return [converted_notebook]


class ConvertNotebookPreprocessorWithFire(ConvertNotebookPreprocessor):
    """Create an entrpoint using pyfire."""
    def __init__(self, #pylint:disable=dangerous-default-value
                 class_name=None,
                 notebook_file=None,
                 notebook_preprocessor=FilterIncludeCell,
                 executable=None,
                 command=["python"],
                 path_prefix=constants.DEFAULT_DEST_PREFIX,
                 output_map=None,
                 overwrite=True):
        """The init function ConvertNotebookPreprocessorWithFire class
        :param class_name: the name of the notebook preprocessor.
        :param notebook_file: the notebook file.
        :param notebook_preprocessor: the class FilterIncludeCell.
        :param command: the python command.
        :param path_prefix: the defaut destion path prefix '/app/'.
        :param output_map: a dict of files to be added without preprocessing.
        """

        super().__init__(
            notebook_file=notebook_file,
            notebook_preprocessor=notebook_preprocessor,
            executable=executable,
            command=command,
            path_prefix=path_prefix,
            output_map=output_map)

        self.class_name = class_name
        self.overwrite = overwrite

    def preprocess(self):
        """Preprocessor the Notebook.
        :return: results: the preprocessed notebook list.
        """
        exporter = nbconvert.PythonExporter()
        exporter.register_preprocessor(self.notebook_preprocessor, enabled=True)
        processed, _ = exporter.from_filename(self.notebook_file)

        lines = []
        for l in processed.splitlines():
            # Get rid of multiple blank lines
            if not l.strip():
                if lines:
                    if not lines[-1]:
                        # last line is already blank don't add another one
                        continue
            # strip in statements
            if l.startswith("# In["):
                continue
            lines.append(l)

        contents = "\n".join(lines)
        converted_notebook = Path(self.notebook_file).with_suffix('.py')
        if converted_notebook.exists() and not self.overwrite:
            raise Exception('Default path {} exists but overwrite flag\
                            is False'.format(converted_notebook))
        with open(converted_notebook, 'w') as f:
            logging.info('Converting {} to {}'.format(self.notebook_file, converted_notebook))
            f.write(contents)
            f.write("\n")
            logging.info('Creating entry point for the class name {}'.format(self.class_name))
            f.write("""
if __name__ == "__main__":
  import fire
  import logging
  logging.basicConfig(format='%(message)s')
  logging.getLogger().setLevel(logging.INFO)
  fire.Fire({0})
""".format(self.class_name))
        self.executable = converted_notebook
        results = [converted_notebook]
        results.extend(self.input_files)
        return results



from googleapiclient import discovery
from googleapiclient import errors

from kubeflow.fairing.deployers.deployer import DeployerInterface
from kubeflow.fairing.cloud.gcp import guess_project_name
from kubeflow.fairing import http_utils


# TODO: Implement predict and delete methods.
class GCPServingDeployer(DeployerInterface):
    """Handle deploying a trained model to GCP."""
    def __init__(self, model_dir, model_name, version_name, project_id=None,
                 **deploy_kwargs):
        self._project_id = project_id or guess_project_name()

        self._model_dir = model_dir
        self._model_name = model_name
        self._version_name = version_name
        self._deploy_kwargs = deploy_kwargs
        self._ml = discovery.build('ml', 'v1')
        self._ml._http = http_utils.configure_http_instance(self._ml._http) #pylint:disable=protected-access

        # Set default deploy kwargs
        if 'runtime_version' not in self._deploy_kwargs:
            self._deploy_kwargs['runtime_version'] = '1.13'
        if 'python_version' not in self._deploy_kwargs:
            self._deploy_kwargs['python_version'] = '3.5'

    def deploy(self, pod_template_spec):
        """Deploys the model to Cloud ML Engine.

        :param pod_template_spec: pod spec template of training job

        """
        # Check if the model exists
        try:
            res = self._ml.projects().models().get(
                name='projects/{}/models/{}'.format(self._project_id, self._model_name)
            ).execute()
        except errors.HttpError as err:
            if err.resp['status'] == '404':
                # Model not found
                res = None
            else:
                # Other error with the command
                print('Error retrieving the model: {}'.format(err))
                return

        if res is None:
            # Create the model
            try:
                model_body = {'name': self._model_name}
                res = self._ml.projects().models().create(
                    parent='projects/{}'.format(self._project_id),
                    body=model_body
                ).execute()
            except errors.HttpError as err:
                print('Error creating the model: {}'.format(err))
                return

        # Create the version
        try:
            version_body = self._deploy_kwargs
            version_body['name'] = self._version_name
            version_body['deploymentUri'] = self._model_dir

            res = self._ml.projects().models().versions().create(
                parent='projects/{}/models/{}'.format(
                    self._project_id, self._model_name),
                body=version_body
            ).execute()
        except errors.HttpError as err:
            print('Error creating the version: {}'.format(err))
            return

        print('Version submitted successfully. Access the version at the following URL:')
        print('https://console.cloud.google.com/mlengine/models/{}/versions/{}?project={}'.format(
            self._model_name, self._version_name, self._project_id))

    def get_logs(self):
        """ abstract get log"""
        raise NotImplementedError('Retrieving logs is not supported for the GCP Serving deployer.')



from kubeflow.fairing.preprocessors.base import BasePreProcessor
from kubeflow.fairing.preprocessors.converted_notebook import ConvertNotebookPreprocessor
from kubeflow.fairing.preprocessors.full_notebook import FullNotebookPreProcessor
from kubeflow.fairing.preprocessors.function import FunctionPreProcessor

from kubeflow.fairing.builders.append.append import AppendBuilder
from kubeflow.fairing.builders.docker.docker import DockerBuilder
from kubeflow.fairing.builders.cluster.cluster import ClusterBuilder

from kubeflow.fairing.deployers.job.job import Job
from kubeflow.fairing.deployers.serving.serving import Serving
from kubeflow.fairing.deployers.tfjob.tfjob import TfJob
from kubeflow.fairing.deployers.gcp.gcp import GCPJob
from kubeflow.fairing.deployers.kfserving.kfserving import KFServing

from kubeflow.fairing.notebook import notebook_util

import logging
logging.basicConfig(format='%(message)s')

DEFAULT_PREPROCESSOR = 'python'
DEFAULT_BUILDER = 'append'
DEFAULT_DEPLOYER = 'job'

preprocessor_map = {
    'python': BasePreProcessor,
    'notebook': ConvertNotebookPreprocessor,
    'full_notebook': FullNotebookPreProcessor,
    'function': FunctionPreProcessor,
}

builder_map = {
    'append': AppendBuilder,
    'docker': DockerBuilder,
    'cluster': ClusterBuilder,
}

deployer_map = {
    'job': Job,
    'tfjob': TfJob,
    'gcp': GCPJob,
    'serving': Serving,
    'kfserving': KFServing,
}


class Config(object):
    def __init__(self):
        self.reset()

    def reset(self):
        """ reset the preprocessor, builder_name and deployer name"""
        if notebook_util.is_in_notebook():
            self._preprocessor_name = 'notebook'
        else:
            self._preprocessor_name = DEFAULT_PREPROCESSOR
        self._preprocessor_kwargs = {}

        self._builder_name = DEFAULT_BUILDER
        self._builder_kwargs = {}

        self._deployer_name = DEFAULT_DEPLOYER
        self._deployer_kwargs = {}

    def set_preprocessor(self, name=None, **kwargs):
        """

        :param name: preprocessor name(Default value = None)

        """
        self._preprocessor_name = name
        self._preprocessor_kwargs = kwargs

    def get_preprocessor(self):
        """ get the preprocessor"""
        fn = preprocessor_map.get(self._preprocessor_name)
        if fn is None:
            raise Exception('Preprocessor name not found: {}\nAvailable preprocessor: {}'.format(
                self._preprocessor_name, list(preprocessor_map.keys())))
        return fn(**self._preprocessor_kwargs)

    def set_builder(self, name=DEFAULT_BUILDER, **kwargs):
        """set the builder

        :param name: builder name (Default value = DEFAULT_BUILDER)

        """
        self._builder_name = name
        self._builder_kwargs = kwargs

    def get_builder(self, preprocessor):
        """get the builder

        :param preprocessor: preprocessor function

        """
        fn = builder_map.get(self._builder_name)
        if fn is None:
            raise Exception('Builder name not found: {}\nAvailable builder: {}'.format(
                self._builder_name, list(builder_map.keys())))
        return fn(preprocessor=preprocessor, **self._builder_kwargs)

    def set_deployer(self, name=DEFAULT_DEPLOYER, **kwargs):
        """set the deployer

        :param name: deployer name (Default value = DEFAULT_DEPLOYER)

        """
        self._deployer_name = name
        self._deployer_kwargs = kwargs

    def get_deployer(self):
        """ get deployer"""
        fn = deployer_map.get(self._deployer_name)
        if fn is None:
            raise Exception('Deployer name not found: {}\nAvailable deployer: {}'.format(
                self._deployer_name, list(deployer_map.keys())))
        return fn(**self._deployer_kwargs)

    def run(self):
        """ run the pipeline for job"""
        preprocessor = self.get_preprocessor()
        logging.info("Using preprocessor: %s", preprocessor)
        builder = self.get_builder(preprocessor)
        logging.info("Using builder: %s", builder)
        deployer = self.get_deployer()
        logging.info("Using deployer: %s", builder)

        builder.build()
        pod_spec = builder.generate_pod_spec()
        deployer.deploy(pod_spec)

        return preprocessor, builder, deployer

    def deploy(self, pod_spec):
        """deploy the job

        :param pod_spec: pod spec of the job

        """
        return self.get_deployer().deploy(pod_spec)

    def fn(self, fn):
        """function

        :param fn: return func that set the preprocessorr and run

        """
        def ret_fn():
            """ return the function"""
            self.set_preprocessor('function', function_obj=fn)
            self.run()
        return ret_fn


config = Config()



from kubeflow.fairing.cloud import aws
from kubeflow.fairing import utils
from kubeflow.fairing.kubernetes.manager import client, KubeManager
from kubeflow.fairing.builders.cluster.context_source import ContextSourceInterface
from kubeflow.fairing.constants import constants


class S3ContextSource(ContextSourceInterface):
    """aws S3 context for docker builder"""
    def __init__(self,
                 aws_account=None,
                 region=None,
                 bucket_name=None):
        self.aws_account = aws_account
        self.manager = KubeManager()
        self.region = region or 'us-east-1'
        self.bucket_name = bucket_name

    def prepare(self, context_filename):  # pylint:disable=arguments-differ
        """

        :param context_filename: context filename

        """
        if self.aws_account is None:
            self.aws_account = aws.guess_account_id()
        self.uploaded_context_url = self.upload_context(context_filename)

    def upload_context(self, context_filename):
        """

        :param context_filename: context filename

        """
        s3_uploader = aws.S3Uploader(self.region)
        context_hash = utils.crc(context_filename)
        bucket_name = self.bucket_name or 'kubeflow-' + \
            self.aws_account + '-' + self.region
        return s3_uploader.upload_to_bucket(bucket_name=bucket_name,
                                            blob_name='fairing_builds/' + context_hash,
                                            file_to_upload=context_filename)

    def cleanup(self):
        pass

    def generate_pod_spec(self, image_name, push):  # pylint:disable=arguments-differ
        args = ["--dockerfile=Dockerfile",
                "--destination=" + image_name,
                "--context=" + self.uploaded_context_url]
        if not push:
            args.append("--no-push")

        return client.V1PodSpec(
            containers=[client.V1Container(name='kaniko',
                                           image=constants.KANIKO_IMAGE,
                                           args=["--dockerfile=Dockerfile",
                                                 "--destination=" + image_name,
                                                 "--context=" + self.uploaded_context_url],
                                           env=[client.V1EnvVar(name='AWS_REGION',
                                                                value=self.region)])],
            restart_policy='Never')



import boto3
import logging
import re
from botocore.exceptions import ClientError
from kubernetes import client

from kubeflow.fairing.constants import constants

logger = logging.getLogger(__name__)

class S3Uploader(object):
    """ For AWS S3 up load """
    def __init__(self, region):
        self.region = region
        self.storage_client = boto3.client('s3', region_name=region)

    def upload_to_bucket(self,
                         blob_name,
                         bucket_name,
                         file_to_upload):
        """Upload a file to an S3 bucket

        :param blob_name: S3 object name
        :param bucket_name: Bucket to upload to
        :param file_to_upload: File to upload

        """
        self.create_bucket_if_not_exists(bucket_name)
        self.storage_client.upload_file(file_to_upload, bucket_name, blob_name)
        return "s3://{}/{}".format(bucket_name, blob_name)

    def create_bucket_if_not_exists(self, bucket_name):
        """Create bucket if this bucket not exists

        :param bucket_name: Bucket name

        """
        try:
            self.storage_client.head_bucket(Bucket=bucket_name)
        except ClientError:
            bucket = {'Bucket': bucket_name}
            if self.region != 'us-east-1':
                bucket['CreateBucketConfiguration'] = {'LocationConstraint': self.region}
            self.storage_client.create_bucket(**bucket)


def guess_account_id():
    """ Get account id """
    account_id = boto3.client('sts').get_caller_identity()["Account"]

    if account_id is None:
        raise Exception('Could not determine account id.')

    return account_id


def add_aws_credentials_if_exists(kube_manager, pod_spec, namespace):
    """add AWS credential

    :param kube_manager: kube manager for handles communication with Kubernetes' client
    :param pod_spec: pod spec like volumes and security context
    :param namespace: The custom resource

    """
    try:
        if kube_manager.secret_exists(constants.AWS_CREDS_SECRET_NAME, namespace):
            add_aws_credentials(kube_manager, pod_spec, namespace)
        else:
            logger.warning("Not able to find aws credentials secret: {}"
                           .format(constants.AWS_CREDS_SECRET_NAME))
    except Exception as e:
        logger.warning("could not check for secret: {}".format(e))


def add_aws_credentials(kube_manager, pod_spec, namespace):
    """add AWS credential

    :param kube_manager: kube manager for handles communication with Kubernetes' client
    :param pod_spec: pod spec like volumes and security context
    :param namespace: The custom resource

    """
    if not kube_manager.secret_exists(constants.AWS_CREDS_SECRET_NAME, namespace):
        raise ValueError('Unable to mount credentials: Secret aws-secret not found in namespace {}'
                         .format(namespace))

    # Set appropriate secrets env to enable kubeflow-user service
    # account.
    env = [
        client.V1EnvVar(
            name='AWS_ACCESS_KEY_ID',
            value_from=client.V1EnvVarSource(
                secret_key_ref=client.V1SecretKeySelector(
                    name=constants.AWS_CREDS_SECRET_NAME,
                    key='AWS_ACCESS_KEY_ID'
                )
            )
        ),
        client.V1EnvVar(
            name='AWS_SECRET_ACCESS_KEY',
            value_from=client.V1EnvVarSource(
                secret_key_ref=client.V1SecretKeySelector(
                    name=constants.AWS_CREDS_SECRET_NAME,
                    key='AWS_SECRET_ACCESS_KEY'
                )
            )
        )]

    if pod_spec.containers[0].env:
        pod_spec.containers[0].env.extend(env)
    else:
        pod_spec.containers[0].env = env


def add_ecr_config(kube_manager, pod_spec, namespace):
    """add secret

    :param kube_manager: kube manager for handles communication with Kubernetes' client
    :param pod_spec: pod spec like volumes and security context
    :param namespace: The custom resource

    """
    if not kube_manager.secret_exists('ecr-config', namespace):
        secret = client.V1Secret(metadata=client.V1ObjectMeta(name='ecr-config'),
                                 string_data={
                                     'config.json': '{"credsStore": "ecr-login"}'
                                 })
        kube_manager.create_secret(namespace, secret)

    volume_mount = client.V1VolumeMount(name='ecr-config',
                                        mount_path='/kaniko/.docker/', read_only=True)

    if pod_spec.containers[0].volume_mounts:
        pod_spec.containers[0].volume_mounts.append(volume_mount)
    else:
        pod_spec.containers[0].volume_mounts = [volume_mount]

    volume = client.V1Volume(name='ecr-config',
                             secret=client.V1SecretVolumeSource(secret_name='ecr-config'))

    if pod_spec.volumes:
        pod_spec.volumes.append(volume)
    else:
        pod_spec.volumes = [volume]


def is_ecr_registry(registry):
    """verify secrte registy

    :param registry: registry

    """
    pattern = r'(.+)\.dkr\.ecr\.(.+)\.amazonaws\.com'
    return bool(re.match(pattern, registry))


def create_ecr_registry(registry, repository):
    """create secret registry

    :param registry: registry
    :param repository: repository name

    """
    registry = registry.split('.')
    registry_id = registry[0]
    region = registry[3]

    ecr_client = boto3.client('ecr', region_name=region)

    try:
        ecr_client.describe_repositories(registryId=registry_id,
                                         repositoryNames=[repository])
    except ClientError:
        ecr_client.create_repository(repositoryName=repository)



import json
import os.path
import ipykernel
import re
from notebook.notebookapp import list_running_servers
import requests
from requests.compat import urljoin


def get_notebook_name(): #pylint:disable=inconsistent-return-statements
    """Return the full path of the jupyter notebook. """
    kernel_id = re.search('kernel-(.*).json',
                          ipykernel.connect.get_connection_file()).group(1)
    servers = list_running_servers()
    for ss in servers:
        response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                params={'token': ss.get('token', '')})
        for nn in json.loads(response.text):
            if nn['kernel']['id'] == kernel_id:
                full_path = nn['notebook']['path']
                return os.path.basename(full_path)


def is_in_notebook():
    """To check is in notenook or not. """

    try:
        ipykernel.get_connection_info()
    # Temporary fix for #84
    # TODO: remove blanket Exception catching after fixing #84
    except Exception:
        return False
    return True



import httplib2
from kubeflow import fairing
from kubeflow.fairing.constants import constants

def configure_http_instance(http=None):
    """Configure http instance to modify the request headers to append or modify user-agent.

    :param http: Body of googleapiclient (Default value = None)
    :returns: object: Configurated http contents.

    """
    if not http:
        http = httplib2.Http()

    request_orig = http.request
    user_agent = constants.DEFAULT_USER_AGENT.format(VERSION=fairing.__version__)
    # Reference:
    # https://github.com/googleapis/google-api-python-client/blob/master/googleapiclient/http.py
    # The closure that will replace 'httplib2.Http.request'.
    def append_ua(headers):
        headers = headers or {}
        if 'user-agent' in headers:
            headers['user-agent'] = user_agent + " " + headers['user-agent']
        else:
            headers['user-agent'] = user_agent
        return headers
    def new_request(*args, **kwargs):
        """Modify the request headers to add the user-agent."""
        if args and len(args) >= 4:
            args = list(args)
            # args is a tuple so assignment is not possible
            args[3] = append_ua(args[3])
            args = tuple(args)
        else:
            kwargs['headers'] = append_ua(kwargs.get('headers'))
        return request_orig(*args, **kwargs)
    http.request = new_request
    return http



from kubernetes import client, config, watch
from kfserving import KFServingClient

from kubeflow.fairing.utils import is_running_in_k8s
from kubeflow.fairing.constants import constants

import logging
logger = logging.getLogger(__name__)

MAX_STREAM_BYTES = 1024


class KubeManager(object):
    """Handles communication with Kubernetes' client."""

    def __init__(self):
        if is_running_in_k8s():
            config.load_incluster_config()
        else:
            config.load_kube_config()

    def create_job(self, namespace, job):
        """Creates a V1Job in the specified namespace.

        :param namespace: The resource
        :param job: Job defination as kubernetes
        :returns: object: Created Job.

        """
        api_instance = client.BatchV1Api()
        return api_instance.create_namespaced_job(namespace, job)

    def create_tf_job(self, namespace, job):
        """Create the provided TFJob in the specified namespace.
        The TFJob version is defined in TF_JOB_VERSION in fairing.constants.
        The version TFJob need to be installed before creating the TFJob.

        :param namespace: The custom resource
        :param job: The JSON schema of the Resource to create
        :returns: object: Created TFJob.

        """
        api_instance = client.CustomObjectsApi()
        try:
            return api_instance.create_namespaced_custom_object(
                constants.TF_JOB_GROUP,
                constants.TF_JOB_VERSION,
                namespace,
                constants.TF_JOB_PLURAL,
                job
            )
        except client.rest.ApiException:
            raise RuntimeError("Failed to create TFJob. Perhaps the CRD TFJob version "
                               "{} in not installed(If you use different version you can pass it "
                               "as ENV variable called "
                               "`TF_JOB_VERSION`)?".format(constants.TF_JOB_VERSION))

    def delete_tf_job(self, name, namespace):
        """Delete the provided TFJob in the specified namespace.

        :param name: The custom object
        :param namespace: The custom resource
        :returns: object: The deleted TFJob.

        """
        api_instance = client.CustomObjectsApi()
        return api_instance.delete_namespaced_custom_object(
            constants.TF_JOB_GROUP,
            constants.TF_JOB_VERSION,
            namespace,
            constants.TF_JOB_PLURAL,
            name,
            client.V1DeleteOptions())

    def create_deployment(self, namespace, deployment):
        """Create an V1Deployment in the specified namespace.

        :param namespace: The custom resource
        :param deployment: Deployment body to create
        :returns: object: Created V1Deployments.

        """
        api_instance = client.AppsV1Api()
        return api_instance.create_namespaced_deployment(namespace, deployment)

    def create_isvc(self, namespace, isvc):
        """Create the provided InferenceService in the specified namespace.

        :param namespace: The custom resource
        :param InferenceService: The InferenceService body
        :returns: object: Created InferenceService.

        """
        KFServing = KFServingClient()
        try:
            created_isvc = KFServing.create(isvc, namespace=namespace)
            isvc_name = created_isvc['metadata']['name']
            isvc_namespace = created_isvc['metadata']['namespace']
            KFServing.get(isvc_name, isvc_namespace, watch=True)
            return created_isvc
        except client.rest.ApiException:
            raise RuntimeError("Failed to create InferenceService. Perhaps the CRD "
                               "InferenceService version {} is not installed? "\
                                   .format(constants.KFSERVING_VERSION))

    def delete_isvc(self, name, namespace):
        """Delete the provided InferenceService in the specified namespace.

        :param name: The custom object
        :param namespace: The custom resource
        :returns: object: The deleted InferenceService.

        """
        KFServing = KFServingClient()
        return KFServing.delete(name, namespace=namespace)

    def delete_job(self, name, namespace):
        """Delete the specified job and related pods.

        :param name: The job name
        :param namespace: The resource
        :returns: object: the deleted job.

        """
        api_instance = client.BatchV1Api()
        api_instance.delete_namespaced_job(
            name,
            namespace,
            client.V1DeleteOptions())

    def delete_deployment(self, name, namespace):
        """Delete an existing model deployment and relinquish all resources associated.

        :param name: The deployment name
        :param namespace: The custom resource
        :returns: obje   deployment.

        """
        api_instance = client.ExtensionsV1beta1Api()
        api_instance.delete_namespaced_deployment(
            name,
            namespace,
            client.V1DeleteOptions())

    def secret_exists(self, name, namespace):
        """Check if the secret exists in the specified namespace.

        :param name: The secret name
        :param namespace: The custom resource.
        :returns: bool: True if the secret exists, otherwise return False.

        """
        secrets = client.CoreV1Api().list_namespaced_secret(namespace)
        secret_names = [secret.metadata.name for secret in secrets.items]
        return name in secret_names

    def create_secret(self, namespace, secret):
        """Create secret in the specified namespace.

        :param namespace: The custom resource
        :param secret: The secret body
        :returns: object: Created secret.

        """
        api_instance = client.CoreV1Api()
        return api_instance.create_namespaced_secret(namespace, secret)

    def get_service_external_endpoint(self, name, namespace, selectors=None): #pylint:disable=inconsistent-return-statements
        """Get the service external endpoint as http://ip_or_hostname:5000/predict.

        :param name: The sevice name
        :param namespace: The custom resource
        :param selectors: A selector to restrict the list of returned objects by their labels.
        :param Defaults: to everything
        :returns: str: the service external endpoint.

        """
        label_selector_str = ', '.join("{}={}".format(k, v) for (k, v) in selectors.items())
        v1 = client.CoreV1Api()
        w = watch.Watch()
        print("Waiting for prediction endpoint to come up...")
        try:
            for event in w.stream(v1.list_namespaced_service,
                                  namespace=namespace,
                                  label_selector=label_selector_str):
                svc = event['object']
                logger.debug("Event: %s %s",
                             event['type'],
                             event['object'])
                ing = svc.status.load_balancer.ingress
                if ing is not None and len(ing) > 0: #pylint:disable=len-as-condition
                    url = "http://{}:5000/predict".format(ing[0].ip or ing[0].hostname)
                    return url
        except ValueError as v:
            logger.error("error getting status for {} {}".format(name, str(v)))
        except client.rest.ApiException as e:
            logger.error("error getting status for {} {}".format(name, str(e)))

    def log(self, name, namespace, selectors=None, container='', follow=True):
        """Get log of the specified pod.

        :param name: The pod name
        :param namespace: The custom resource
        :param selectors: A selector to restrict the list of returned objects by their labels.
        :param Defaults: to everything
        :param container: The container for which to stream logs.
        :param if: there is one container in the pod
        :param follow: True or False (Default value = True)
        :returns: str: logs of the specified pod.

        """
        tail = ''
        label_selector_str = ', '.join("{}={}".format(k, v) for (k, v) in selectors.items())
        v1 = client.CoreV1Api()
        # Retry to allow starting of pod
        w = watch.Watch()
        try:
            for event in w.stream(v1.list_namespaced_pod,
                                  namespace=namespace,
                                  label_selector=label_selector_str):
                pod = event['object']
                logger.debug("Event: %s %s %s",
                             event['type'],
                             pod.metadata.name,
                             pod.status.phase)
                if pod.status.phase == 'Pending':
                    logger.warning('Waiting for {} to start...'.format(pod.metadata.name))
                    continue
                elif ((pod.status.phase == 'Running'
                       and pod.status.container_statuses[0].ready)
                      or pod.status.phase == 'Succeeded'):
                    logger.info("Pod started running %s",
                                pod.status.container_statuses[0].ready)
                    tail = v1.read_namespaced_pod_log(pod.metadata.name,
                                                      namespace,
                                                      follow=follow,
                                                      _preload_content=False,
                                                      pretty='pretty',
                                                      container=container)
                    break
                elif (event['type'] == 'DELETED'
                      or pod.status.phase == 'Failed'
                      or pod.status.container_statuses[0].state.waiting):
                    logger.error("Failed to launch %s, reason: %s, message: %s",
                                 pod.metadata.name,
                                 pod.status.container_statuses[0].state.terminated.reason,
                                 pod.status.container_statuses[0].state.terminated.message)
                    tail = v1.read_namespaced_pod_log(pod.metadata.name,
                                                      namespace,
                                                      follow=follow,
                                                      _preload_content=False,
                                                      pretty='pretty',
                                                      container=container)
                    break
        except ValueError as v:
            logger.error("error getting status for {} {}".format(name, str(v)))
        except client.rest.ApiException as e:
            logger.error("error getting status for {} {}".format(name, str(e)))
        if tail:
            try:
                for chunk in tail.stream(MAX_STREAM_BYTES):
                    print(chunk.rstrip().decode('utf8'))
            finally:
                tail.release_conn()



import logging
import base64
import tarfile
from pathlib import Path
from shutil import rmtree

from azure.common.credentials import ServicePrincipalCredentials
from azure.mgmt.storage import StorageManagementClient
from azure.mgmt.storage.models import StorageAccountCreateParameters
from azure.mgmt.storage.models import Sku
from azure.mgmt.storage.models import SkuName
from azure.mgmt.storage.models import Kind
from azure.storage.file import FileService
from kubernetes import client

from kubeflow.fairing.constants import constants
from kubeflow.fairing.kubernetes.manager import KubeManager

logger = logging.getLogger(__name__)

# Helper class to upload files to Azure Files
class AzureFileUploader(object):
    def __init__(self, namespace, credentials=None, subscription_id=None):
        if not credentials or not subscription_id:
            credentials, subscription_id = get_azure_credentials(namespace)
        self.storage_client = StorageManagementClient(credentials, subscription_id)

    # Upload the files and dirs in a tar.gz file to a dir in a shared folder in Azure Files
    def upload_to_share(self,
                        region,
                        resource_group_name,
                        storage_account_name,
                        share_name,
                        dir_name,
                        tar_gz_file_to_upload):

        logging.info(
            "Uploading contents of '{}' to 'https://{}.file.core.windows.net/{}/{}'"
            .format(tar_gz_file_to_upload, storage_account_name, share_name, dir_name)
        )

        self.create_storage_account_if_not_exists(region, resource_group_name, storage_account_name)
        storage_account_name, storage_key = self.get_storage_credentials(
            resource_group_name, storage_account_name
        )
        share_service = FileService(account_name=storage_account_name, account_key=storage_key)
        self.create_share_if_not_exists(share_service, share_name)
        share_service.create_directory(share_name, dir_name)
        self.upload_tar_gz_contents(share_service, share_name, dir_name, tar_gz_file_to_upload)

        return storage_account_name, storage_key

    def create_storage_account_if_not_exists(self, region, resource_group_name,
                                             storage_account_name):
        """Creates the storage account if it does not exist.

        In either case, returns the StorageAccount class that matches the given arguments."""
        storage_accounts = (
            self.storage_client.storage_accounts
            .list_by_resource_group(resource_group_name)
        )
        storage_account = next(
            filter(lambda storage_account:
                   storage_account.name == storage_account_name,
                   storage_accounts),
            None
        )
        if storage_account:
            return storage_account
        logging.info(
            "Creating Azure Storage account '{}' in Resource Group '{}'"
            .format(storage_account_name, resource_group_name)
        )
        storage_async_operation = self.storage_client.storage_accounts.create(
            resource_group_name,
            storage_account_name,
            StorageAccountCreateParameters(
                sku=Sku(name=SkuName.standard_ragrs),
                kind=Kind.storage,
                location=region
            )
        )
        return storage_async_operation.result()

    def get_storage_credentials(self, resource_group_name, storage_account_name):
        storage_keys = (
            self.storage_client.storage_accounts
            .list_keys(resource_group_name, storage_account_name)
        )
        storage_keys = {v.key_name: v.value for v in storage_keys.keys}
        return storage_account_name, storage_keys['key1']

    def create_share_if_not_exists(self, share_service, share_name):
        shares = share_service.list_shares()
        share = next(filter(lambda share: share.name == share_name, shares), None)
        if share is None:
            share_service.create_share(share_name)

    def upload_tar_gz_contents(self, share_service, share_name, dir_name, tar_gz_file):
        local_dir = Path('{}_contents'.format(tar_gz_file))
        cloud_dir = Path(dir_name)
        self.uncompress_tar_gz_file(tar_gz_file, local_dir)

        for path in local_dir.glob('**/*'):
            local_path = Path(path)
            cloud_relative_path = cloud_dir / path.relative_to(local_dir)

            if local_path.is_dir():
                share_service.create_directory(share_name, cloud_relative_path)
            else:
                share_service.create_file_from_path(
                    share_name, cloud_relative_path.parents[0],
                    cloud_relative_path.name, local_path
                )

        self.delete_uncompressed_files(local_dir)

    def uncompress_tar_gz_file(self, tar_gz_file, target_dir):
        tar = tarfile.open(tar_gz_file, 'r:gz')
        tar.extractall(path=target_dir)
        tar.close()

    def delete_uncompressed_files(self, target_dir):
        rmtree(target_dir)

# Get credentials for a service principal which has permissions to
# create or access the storage account for Azure Files
def get_azure_credentials(namespace):
    secret_name = constants.AZURE_CREDS_SECRET_NAME
    if not KubeManager().secret_exists(secret_name, namespace):
        raise Exception("Secret '{}' not found in namespace '{}'".format(secret_name, namespace))

    v1 = client.CoreV1Api()
    secret = v1.read_namespaced_secret(secret_name, namespace)
    sp_credentials = ServicePrincipalCredentials(
        client_id=get_plain_secret_value(secret.data, 'AZ_CLIENT_ID'),
        secret=get_plain_secret_value(secret.data, 'AZ_CLIENT_SECRET'),
        tenant=get_plain_secret_value(secret.data, 'AZ_TENANT_ID')
    )
    subscription_id = get_plain_secret_value(secret.data, 'AZ_SUBSCRIPTION_ID')
    return sp_credentials, subscription_id

# Decode plain text value of a secret of given key and raise an exception if the key is not found
def get_plain_secret_value(secret_data, key):
    if not key in secret_data:
        raise Exception("Secret with key '{}'' not found".format(key))
    secret_base64 = secret_data[key]
    return base64.b64decode(secret_base64).decode('utf-8')

# Create a secret with the credentials to access the storage account for Azure Files
def create_storage_creds_secret(namespace, context_hash, storage_account_name, storage_key):
    secret_name = constants.AZURE_STORAGE_CREDS_SECRET_NAME_PREFIX + context_hash.lower()
    logging.info(
        "Creating secret '{}' in namespace '{}'"
        .format(secret_name, namespace)
    )
    secret = client.V1Secret(
        metadata=client.V1ObjectMeta(name=secret_name),
        string_data={
            'azurestorageaccountname': storage_account_name,
            'azurestorageaccountkey': storage_key
        })
    v1 = client.CoreV1Api()
    v1.create_namespaced_secret(namespace, secret)

# Delete the secret with the credentials to access the storage account for Azure Files
def delete_storage_creds_secret(namespace, context_hash):
    secret_name = constants.AZURE_STORAGE_CREDS_SECRET_NAME_PREFIX + context_hash.lower()
    logging.info(
        "Deleting secret '{}' from namespace '{}'"
        .format(secret_name, namespace)
    )
    v1 = client.CoreV1Api()
    v1.delete_namespaced_secret(secret_name, namespace, body=None)

# Verify that we are working with an Azure Container Registry
def is_acr_registry(registry):
    return registry.endswith('.azurecr.io')

# Mount Docker config so the pod can access Azure Container Registry
def add_acr_config(kube_manager, pod_spec, namespace):
    secret_name = constants.AZURE_ACR_CREDS_SECRET_NAME
    if not kube_manager.secret_exists(secret_name, namespace):
        raise Exception("Secret '{}' not found in namespace '{}'".format(secret_name, namespace))

    volume_mount = client.V1VolumeMount(
        name='acr-config', mount_path='/kaniko/.docker/', read_only=True
    )

    if pod_spec.containers[0].volume_mounts:
        pod_spec.containers[0].volume_mounts.append(volume_mount)
    else:
        pod_spec.containers[0].volume_mounts = [volume_mount]

    items = [client.V1KeyToPath(key='.dockerconfigjson', path='config.json')]
    volume = client.V1Volume(
        name='acr-config',
        secret=client.V1SecretVolumeSource(secret_name=secret_name, items=items)
    )
    if pod_spec.volumes:
        pod_spec.volumes.append(volume)
    else:
        pod_spec.volumes = [volume]

# Mount Azure Files shared folder so the pod can access its files with a local path
def add_azure_files(kube_manager, pod_spec, namespace):
    context_hash = pod_spec.containers[0].args[1].split(':')[-1]
    secret_name = constants.AZURE_STORAGE_CREDS_SECRET_NAME_PREFIX + context_hash.lower()
    if not kube_manager.secret_exists(secret_name, namespace):
        raise Exception("Secret '{}' not found in namespace '{}'".format(secret_name, namespace))

    volume_mount = client.V1VolumeMount(
        name='azure-files', mount_path='/mnt/azure/', read_only=True
    )

    if pod_spec.containers[0].volume_mounts:
        pod_spec.containers[0].volume_mounts.append(volume_mount)
    else:
        pod_spec.containers[0].volume_mounts = [volume_mount]

    volume = client.V1Volume(
        name='azure-files',
        azure_file=client.V1AzureFileVolumeSource(
            secret_name=secret_name, share_name=constants.AZURE_FILES_SHARED_FOLDER
        )
    )

    if pod_spec.volumes:
        pod_spec.volumes.append(volume)
    else:
        pod_spec.volumes = [volume]



from kubernetes import client
from kubernetes.client.models.v1_resource_requirements import V1ResourceRequirements
from kubeflow.fairing.constants import constants


def get_resource_mutator(cpu=None, memory=None):
    """The mutator for getting the resource setting for pod spec.

    The useful example:
    https://github.com/kubeflow/fairing/blob/master/examples/train_job_api/main.ipynb

    :param cpu: Limits and requests for CPU resources (Default value = None)
    :param memory: Limits and requests for memory (Default value = None)
    :returns: object: The mutator function for setting cpu and memory in pod spec.

    """
    def _resource_mutator(kube_manager, pod_spec, namespace): #pylint:disable=unused-argument
        if cpu is None and memory is None:
            return
        if pod_spec.containers and len(pod_spec.containers) >= 1:
            # All cloud providers specify their instace memory in GB
            # so it is peferable for user to specify memory in GB
            # and we convert it to Gi that K8s needs
            limits = {}
            if cpu:
                limits['cpu'] = cpu
            if memory:
                memory_gib = "{}Gi".format(round(memory/1.073741824, 2))
                limits['memory'] = memory_gib
            if pod_spec.containers[0].resources:
                if pod_spec.containers[0].resources.limits:
                    pod_spec.containers[0].resources.limits = {}
                for k, v in limits.items():
                    pod_spec.containers[0].resources.limits[k] = v
            else:
                pod_spec.containers[0].resources = V1ResourceRequirements(limits=limits)
    return _resource_mutator


def mounting_pvc(pvc_name, pvc_mount_path=constants.PVC_DEFAULT_MOUNT_PATH):
    """The function for pod_spec_mutators to mount persistent volume claim.

    :param pvc_name: The name of persistent volume claim
    :param pvc_mount_path: Path for the persistent volume claim mounts to.
    :returns: object: function for mount the pvc to pods.

    """
    mounting_name = str(constants.PVC_DEFAULT_VOLUME_NAME) + pvc_name
    def _mounting_pvc(kube_manager, pod_spec, namespace): #pylint:disable=unused-argument
        volume_mount = client.V1VolumeMount(
            name=mounting_name, mount_path=pvc_mount_path)
        if pod_spec.containers[0].volume_mounts:
            pod_spec.containers[0].volume_mounts.append(volume_mount)
        else:
            pod_spec.containers[0].volume_mounts = [volume_mount]

        volume = client.V1Volume(
            name=mounting_name,
            persistent_volume_claim=client.V1PersistentVolumeClaimVolumeSource(claim_name=pvc_name))
        if pod_spec.volumes:
            pod_spec.volumes.append(volume)
        else:
            pod_spec.volumes = [volume]
    return _mounting_pvc



import os
import tarfile
import logging
import posixpath
import tempfile

from kubeflow import fairing
from kubeflow.fairing.constants import constants
from kubeflow.fairing import utils

class BasePreProcessor(object):
    """Prepares a context that gets sent to the builder for the docker build and sets the entrypoint
    :param input_files: the source files to be processed
    :param executable: the file to execute using command (e.g. main.py)
    :param output_map: a dict of files to be added without preprocessing
    :param path_prefix: the prefix of the path where the files will be added in the container
    :param command: the command to pass to the builder

    """
    def __init__(self,
                 input_files=None,
                 command=None,
                 executable=None,
                 path_prefix=constants.DEFAULT_DEST_PREFIX,
                 output_map=None):
        self.executable = executable
        input_files = input_files or []
        command = command or ["python"]
        self.input_files = set([os.path.normpath(f) for f in input_files])
        output_map = output_map if output_map else {}
        normalized_map = {}
        for src, dst in output_map.items():
            normalized_map[os.path.normpath(src)] = os.path.normpath(dst)
        self.output_map = normalized_map

        self.path_prefix = path_prefix
        self.command = command

        self.set_default_executable()

    # TODO: Add workaround for users who do not want to set an executable for
    # their command.
    def set_default_executable(self): #pylint:disable=inconsistent-return-statements
        """ Set the default executable file.

        :returns: executable: get the default executable file if it is not existing, Or None

        """
        if self.executable is not None:
            return self.executable
        if len(self.input_files) == 1:
            self.executable = list(self.input_files)[0]
            return
        python_files = [item for item in self.input_files if item.endswith(".py")
                        and item is not '__init__.py'] #pylint:disable=literal-comparison
        if len(python_files) == 1:
            self.executable = python_files[0]
            return

    def preprocess(self):
        """ Preprocess the 'input_files'.

        :returns: input_files: get the input files

        """
        return self.input_files

    def context_map(self):
        """ Create context mapping from destination to source to avoid duplicates in context archive

        :returns: c_map: a context map

        """
        c_map = {}
        for src, dst in self.fairing_runtime_files().items():
            if dst not in c_map:
                c_map[dst] = src
            else:
                logging.warning('{} already exists in Fairing context, skipping...'.format(src))

        for f in self.input_files:
            dst = os.path.join(self.path_prefix, f)
            if dst not in c_map:
                c_map[dst] = f
            else:
                logging.warning('{} already exists in Fairing context, skipping...'.format(f))

        for src, dst in self.output_map.items():
            if dst not in c_map:
                c_map[dst] = src
            else:
                logging.warning('{} already exists in Fairing context, skipping...'.format(src))

        return c_map

    def context_tar_gz(self, output_file=None):
        """Creating docker context file and compute a running cyclic redundancy check checksum.

        :param output_file: output file (Default value = None)
        :returns: output_file,checksum: docker context file and checksum

        """
        if not output_file:
            _, output_file = tempfile.mkstemp(prefix="/tmp/fairing_context_")
        logging.info("Creating docker context: %s", output_file)
        self.input_files = self.preprocess()
        with tarfile.open(output_file, "w:gz", dereference=True) as tar:
            for dst, src in self.context_map().items():
                logging.debug("Context: %s, Adding %s at %s", output_file,
                              src, dst)
                tar.add(src, filter=reset_tar_mtime, arcname=dst, recursive=False)
        self._context_tar_path = output_file
        return output_file, utils.crc(self._context_tar_path)

    def get_command(self):
        """ Get the execute with absolute path

        :returns: cmd: the execute with absolute path
        """
        if self.command is None:
            return []
        cmd = self.command.copy()
        if self.executable is not None:
            cmd.append(os.path.join(self.path_prefix, self.executable))
        return cmd

    def fairing_runtime_files(self):
        """Search the fairing runtime files 'runtime_config.py'
        :returns: cmd: the execute with absolute path
        """
        fairing_dir = os.path.dirname(fairing.__file__)
        ret = {}
        for f in ["__init__.py", "runtime_config.py"]:
            src = os.path.normpath(os.path.join(fairing_dir, f))
            dst = os.path.normpath(os.path.join(self.path_prefix, "fairing", f))
            ret[src] = dst
        return ret

    def is_requirements_txt_file_present(self):
        """ Verfiy the requirements txt file if it is present.

        :returns: res: get the present required files

        """
        dst_files = self.context_map().keys()
        reqs_file = posixpath.join(self.path_prefix, "requirements.txt")
        res = reqs_file in dst_files
        return res

def reset_tar_mtime(tarinfo):
    """Reset the mtime on the the tarball for reproducibility.

    :param tarinfo: the tarball var
    :returns: tarinfo: the modified tar ball
    """
    tarinfo.mtime = 0
    return tarinfo



import os

from kubeflow.fairing.cloud import gcp
from kubeflow.fairing import utils
from kubeflow.fairing.constants import constants
from kubeflow.fairing.kubernetes.manager import client, KubeManager
from kubeflow.fairing.builders.cluster.context_source import ContextSourceInterface

class GCSContextSource(ContextSourceInterface):
    """Google cloud storage context for docker builder"""
    def __init__(self,
                 gcp_project=None,
                 credentials_file=os.environ.get(constants.GOOGLE_CREDS_ENV),
                 namespace='default'):
        self.gcp_project = gcp_project
        self.credentials_file = credentials_file
        self.manager = KubeManager()
        self.namespace = namespace

    def prepare(self, context_filename):  # pylint:disable=arguments-differ
        if self.gcp_project is None:
            self.gcp_project = gcp.guess_project_name()
        self.uploaded_context_url = self.upload_context(context_filename)

    def upload_context(self, context_filename):
        gcs_uploader = gcp.GCSUploader()
        context_hash = utils.crc(context_filename)
        return gcs_uploader.upload_to_bucket(bucket_name=self.gcp_project,
                                             blob_name='fairing_builds/' + context_hash,
                                             file_to_upload=context_filename)

    def cleanup(self):
        pass

    def generate_pod_spec(self, image_name, push):  # pylint:disable=arguments-differ
        args = ["--dockerfile=Dockerfile",
                "--destination=" + image_name,
                "--context=" + self.uploaded_context_url,
                "--cache=true"]
        if not push:
            args.append("--no-push")

        return client.V1PodSpec(
            containers=[client.V1Container(
                name='kaniko',
                image=constants.KANIKO_IMAGE,
                args=args,
            )],
            restart_policy='Never'
        )



import logging
import json
import uuid

from kubernetes import client as k8s_client

from kubeflow.fairing import utils
from kubeflow.fairing.constants import constants
from kubeflow.fairing.kubernetes.manager import KubeManager
from kubeflow.fairing.deployers.deployer import DeployerInterface


logger = logging.getLogger(__name__)


class Job(DeployerInterface): #pylint:disable=too-many-instance-attributes
    """Handle all the k8s' template building for a training"""

    def __init__(self, namespace=None, runs=1, output=None,
                 cleanup=True, labels=None, job_name=constants.JOB_DEFAULT_NAME,
                 stream_log=True, deployer_type=constants.JOB_DEPLOPYER_TYPE,
                 pod_spec_mutators=None, annotations=None):
        """

        :param namespace: k8s namespace where the training's components will be deployed.
        :param runs: Number of training(s) to be deployed. Hyperparameter search
                will generate multiple jobs.
        :param output: output
        :param cleanup: clean up deletes components after job finished
        :param labels: labels to be assigned to the training job
        :param job_name: name of the job
        :param stream_log: stream the log?
        :param deployer_type: type of deployer
        :param pod_spec_mutators: pod spec mutators (Default value = None)
        """
        if namespace is None:
            self.namespace = utils.get_default_target_namespace()
        else:
            self.namespace = namespace

        # Used as pod and job name
        self.job_name = job_name
        self.deployer_type = deployer_type
        self.deployment_spec = None
        self.runs = runs
        self.output = output
        self.backend = KubeManager()
        self.cleanup = cleanup
        self.stream_log = stream_log
        self.set_labels(labels, deployer_type)
        self.set_anotations(annotations)
        self.pod_spec_mutators = pod_spec_mutators or []

    def set_anotations(self, annotations):
        self.annotations = {}
        if annotations:
            self.annotations.update(annotations)

    def set_labels(self, labels, deployer_type):
        """set labels for the pods of a deployed job

        :param labels: dictionary of labels {label_name:label_value}
        :param deployer_type: deployer type name

        """
        self.labels = {'fairing-deployer': deployer_type}
        if labels:
            self.labels.update(labels)

    def deploy(self, pod_spec): #pylint:disable=arguments-differ
        """deploy the training job using k8s client lib

        :param pod_spec: pod spec of deployed training job

        """
        self.job_id = str(uuid.uuid1())
        self.labels['fairing-id'] = self.job_id
        for fn in self.pod_spec_mutators:
            fn(self.backend, pod_spec, self.namespace)
        pod_template_spec = self.generate_pod_template_spec(pod_spec)
        pod_template_spec.spec.restart_policy = 'Never'
        pod_template_spec.spec.containers[0].name = 'fairing-job'
        self.deployment_spec = self.generate_deployment_spec(pod_template_spec)
        if self.output:
            api = k8s_client.ApiClient()
            job_output = api.sanitize_for_serialization(self.deployment_spec)
            print(json.dumps(job_output))

        name = self.create_resource()
        logger.warning("The {} {} launched.".format(self.deployer_type, name))

        if self.stream_log:
            self.get_logs()

        return name

    def create_resource(self):
        """ create job"""
        self._created_job = self.backend.create_job(self.namespace, self.deployment_spec)
        return self._created_job.metadata.name

    def generate_pod_template_spec(self, pod_spec):
        """Generate a V1PodTemplateSpec initiazlied with correct metadata
            and with the provided pod_spec

        :param pod_spec: pod spec

        """
        if not isinstance(pod_spec, k8s_client.V1PodSpec):
            raise TypeError('pod_spec must be a V1PodSpec, but got %s'
                            % type(pod_spec))
        if not self.annotations:
            self.annotations = {'sidecar.istio.io/inject': 'false'}
        else:
            self.annotations['sidecar.istio.io/inject'] = 'false'
        return k8s_client.V1PodTemplateSpec(
            metadata=k8s_client.V1ObjectMeta(name="fairing-deployer",
                                             annotations=self.annotations,
                                             labels=self.labels),
            spec=pod_spec)

    def generate_deployment_spec(self, pod_template_spec):
        """Generate a V1Job initialized with correct completion and
         parallelism (for HP search) and with the provided V1PodTemplateSpec

        :param pod_template_spec: V1PodTemplateSpec

        """
        if not isinstance(pod_template_spec, k8s_client.V1PodTemplateSpec):
            raise TypeError("""pod_template_spec must be a V1PodTemplateSpec,
                but got %s""" % type(pod_template_spec))

        job_spec = k8s_client.V1JobSpec(
            template=pod_template_spec,
            parallelism=self.runs,
            completions=self.runs,
            backoff_limit=0,
        )

        return k8s_client.V1Job(
            api_version="batch/v1",
            kind="Job",
            metadata=k8s_client.V1ObjectMeta(
                generate_name=self.job_name,
                labels=self.labels,
            ),
            spec=job_spec
        )

    def get_logs(self):
        """ get logs from the deployed job"""
        self.backend.log(self._created_job.metadata.name,
                         self._created_job.metadata.namespace,
                         self.labels,
                         container="fairing-job")

        if self.cleanup:
            self.do_cleanup()

    def do_cleanup(self):
        """ clean up the pods after job finished"""
        logger.warning("Cleaning up job {}...".format(self._created_job.metadata.name))
        k8s_client.BatchV1Api().delete_namespaced_job(
            self._created_job.metadata.name,
            self._created_job.metadata.namespace,
            body=k8s_client.V1DeleteOptions(propagation_policy='Foreground'))



import abc
import six

@six.add_metaclass(abc.ABCMeta)
class BuilderInterface(object):

    @abc.abstractmethod
    def build(self):
        """Will be called when the build needs to start"""
        raise NotImplementedError('BuilderInterface.build')

    @abc.abstractmethod
    def generate_pod_spec(self):
        """This method should return a V1PodSpec with the correct image set.
            This is also where the builder should set the environment variables
            and volume/volumeMounts that it may need to work

        """
        raise NotImplementedError('BuilderInterface.generate_pod_spec')



import logging

from kubernetes import client

from kubeflow.fairing.builders.builder import BuilderInterface
from kubeflow.fairing.constants import constants
from kubeflow.fairing.cloud import gcp

logger = logging.getLogger(__name__)


class BaseBuilder(BuilderInterface): #pylint:disable=too-many-instance-attributes
    """A builder using the local Docker client"""

    def __init__(self,
                 registry=None,
                 image_name=None,
                 base_image=constants.DEFAULT_BASE_IMAGE,
                 push=True,
                 preprocessor=None,
                 dockerfile_path=None):

        self.registry = registry
        self.image_name = image_name
        self.push = push
        if self.registry is None:
            # TODO(r2d4): Add more heuristics here...

            # If no push and no registry provided, use any registry name
            if not self.push:
                self.registry = 'local/fairing-image'
            else:
                self.registry = 'gcr.io/{}'.format(gcp.guess_project_name())

        self.base_image = base_image
        self.dockerfile_path = dockerfile_path
        self.preprocessor = preprocessor
        self.image_tag = None
        self.docker_client = None

    def generate_pod_spec(self):
        return client.V1PodSpec(
            containers=[client.V1Container(
                name='model',
                image=self.image_tag,
                command=self.preprocessor.get_command(),
                security_context=client.V1SecurityContext(
                    run_as_user=0,
                ),
                env=[client.V1EnvVar(
                    name='FAIRING_RUNTIME',
                    value='1',
                )],
                # Set the directory where the python files are built.
                # TODO(jlewi): Would it be better to set PYTHONPATH?
                working_dir=self.preprocessor.path_prefix,
            )],
        )

    def full_image_name(self, tag):
        """Retrun the full image name

        :param tag:  the new tag for the image

        """
        return '{}/{}:{}'.format(self.registry, self.image_name, tag)

    def build(self):
        """Runs the build"""
        raise NotImplementedError()



import abc
import six


@six.add_metaclass(abc.ABCMeta)
class ContextSourceInterface(object):
    """Interface that provides the build context to the in cluster builder"""

    @abc.abstractmethod
    def prepare(self):
        """Makes the context somehow available for use in the pod spec"""
        raise NotImplementedError('ContextSourceInterface.setup')

    @abc.abstractmethod
    def cleanup(self):
        """Cleans up the context after the build is complete"""
        raise NotImplementedError('ContextSourceInterface.cleanup')

    @abc.abstractmethod
    def generate_pod_spec(self, pod_spec):
        """Generates a pod spec for building the image in the cluster, pointing to
        the prepared build context

        :param pod_spec: pod spec

        """
        raise NotImplementedError('ContextSourceInterface.modify_pod_spec')



import logging
import uuid

from kubernetes import client

from kubeflow.fairing import utils
from kubeflow.fairing.builders.base_builder import BaseBuilder
from kubeflow.fairing.builders import dockerfile
from kubeflow.fairing.constants import constants
from kubeflow.fairing.kubernetes.manager import KubeManager

logger = logging.getLogger(__name__)


class ClusterBuilder(BaseBuilder):
    """Builds a docker image in a Kubernetes cluster.
    """

    def __init__(self,
                 registry=None,
                 image_name=constants.DEFAULT_IMAGE_NAME,
                 context_source=None,
                 preprocessor=None,
                 push=True,
                 base_image=constants.DEFAULT_BASE_IMAGE,
                 pod_spec_mutators=None,
                 namespace=None,
                 dockerfile_path=None,
                 cleanup=False):
        super().__init__(
            registry=registry,
            image_name=image_name,
            push=push,
            preprocessor=preprocessor,
            base_image=base_image,
            dockerfile_path=dockerfile_path)
        self.manager = KubeManager()
        if context_source is None:
            raise RuntimeError("context_source is not specified")
        self.context_source = context_source
        self.pod_spec_mutators = pod_spec_mutators or []
        self.namespace = namespace or utils.get_default_target_namespace()
        self.cleanup = cleanup

    def build(self):
        logging.info("Building image using cluster builder.")
        install_reqs_before_copy = self.preprocessor.is_requirements_txt_file_present()
        if self.dockerfile_path:
            dockerfile_path = self.dockerfile_path
        else:
            dockerfile_path = dockerfile.write_dockerfile(
                path_prefix=self.preprocessor.path_prefix,
                base_image=self.base_image,
                install_reqs_before_copy=install_reqs_before_copy
            )
        self.preprocessor.output_map[dockerfile_path] = 'Dockerfile'
        context_path, context_hash = self.preprocessor.context_tar_gz()
        self.image_tag = self.full_image_name(context_hash)
        self.context_source.prepare(context_path)
        labels = {'fairing-builder': 'kaniko'}
        labels['fairing-build-id'] = str(uuid.uuid1())
        pod_spec = self.context_source.generate_pod_spec(
            self.image_tag, self.push)
        for fn in self.pod_spec_mutators:
            fn(self.manager, pod_spec, self.namespace)

        pod_spec_template = client.V1PodTemplateSpec(
            metadata=client.V1ObjectMeta(
                generate_name="fairing-builder-",
                labels=labels,
                namespace=self.namespace,
                annotations={"sidecar.istio.io/inject": "false"},
            ),
            spec=pod_spec
        )
        job_spec = client.V1JobSpec(
            template=pod_spec_template,
            parallelism=1,
            completions=1,
            backoff_limit=0,
        )
        build_job = client.V1Job(
            api_version="batch/v1",
            kind="Job",
            metadata=client.V1ObjectMeta(
                generate_name="fairing-builder-",
                labels=labels,
            ),
            spec=job_spec
        )
        created_job = client. \
            BatchV1Api(). \
            create_namespaced_job(self.namespace, build_job)

        self.manager.log(
            name=created_job.metadata.name,
            namespace=created_job.metadata.namespace,
            selectors=labels,
            container="kaniko")

        # Invoke upstream clean ups
        self.context_source.cleanup()
        # Cleanup build_job if requested by user
        # Otherwise build_job will be cleaned up by Kubernetes GC
        if self.cleanup:
            logging.warning("Cleaning up job {}...".format(created_job.metadata.name))
            client. \
                BatchV1Api(). \
                delete_namespaced_job(
                    created_job.metadata.name,
                    created_job.metadata.namespace,
                    body=client.V1DeleteOptions(propagation_policy='Foreground')
                )



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This package provides DockerImage for examining docker_build outputs."""

from __future__ import absolute_import
from __future__ import division

from __future__ import print_function

import abc
import gzip
import io
import json
import os
import tarfile
import threading

from containerregistry.client import docker_creds
from containerregistry.client import docker_name
from containerregistry.client.v2_2 import docker_digest
from containerregistry.client.v2_2 import docker_http
import httplib2
import six
from six.moves import zip  # pylint: disable=redefined-builtin
import six.moves.http_client


class DigestMismatchedError(Exception):
  """Exception raised when a digest mismatch is encountered."""


class DockerImage(six.with_metaclass(abc.ABCMeta, object)):
  """Interface for implementations that interact with Docker images."""

  def fs_layers(self):
    """The ordered collection of filesystem layers that comprise this image."""
    manifest = json.loads(self.manifest())
    return [x['digest'] for x in reversed(manifest['layers'])]

  def diff_ids(self):
    """The ordered list of uncompressed layer hashes (matches fs_layers)."""
    cfg = json.loads(self.config_file())
    return list(reversed(cfg.get('rootfs', {}).get('diff_ids', [])))

  def config_blob(self):
    manifest = json.loads(self.manifest())
    return manifest['config']['digest']

  def blob_set(self):
    """The unique set of blobs that compose to create the filesystem."""
    return set(self.fs_layers() + [self.config_blob()])

  def distributable_blob_set(self):
    """The unique set of blobs which are distributable."""
    manifest = json.loads(self.manifest())
    distributable_blobs = {
        x['digest']
        for x in reversed(manifest['layers'])
        if x['mediaType'] not in docker_http.NON_DISTRIBUTABLE_LAYER_MIMES
    }
    distributable_blobs.add(self.config_blob())
    return distributable_blobs

  def digest(self):
    """The digest of the manifest."""
    return docker_digest.SHA256(self.manifest().encode('utf8'))

  def media_type(self):
    """The media type of the manifest."""
    manifest = json.loads(self.manifest())
    # Since 'mediaType' is optional for OCI images, assume OCI if it's missing.
    return manifest.get('mediaType', docker_http.OCI_MANIFEST_MIME)

  # pytype: disable=bad-return-type
  @abc.abstractmethod
  def manifest(self):
    """The JSON manifest referenced by the tag/digest.

    Returns:
      The raw json manifest
    """
  # pytype: enable=bad-return-type

  # pytype: disable=bad-return-type
  @abc.abstractmethod
  def config_file(self):
    """The raw blob bytes of the config file."""
  # pytype: enable=bad-return-type

  def blob_size(self, digest):
    """The byte size of the raw blob."""
    return len(self.blob(digest))

  # pytype: disable=bad-return-type
  @abc.abstractmethod
  def blob(self, digest):
    """The raw blob of the layer.

    Args:
      digest: the 'algo:digest' of the layer being addressed.

    Returns:
      The raw blob bytes of the layer.
    """
  # pytype: enable=bad-return-type

  def uncompressed_blob(self, digest):
    """Same as blob() but uncompressed."""
    zipped = self.blob(digest)
    buf = io.BytesIO(zipped)
    f = gzip.GzipFile(mode='rb', fileobj=buf)
    unzipped = f.read()
    return unzipped

  def _diff_id_to_digest(self, diff_id):
    for (this_digest, this_diff_id) in six.moves.zip(self.fs_layers(),
                                                     self.diff_ids()):
      if this_diff_id == diff_id:
        return this_digest
    raise ValueError('Unmatched "diff_id": "%s"' % diff_id)

  def digest_to_diff_id(self, digest):
    for (this_digest, this_diff_id) in six.moves.zip(self.fs_layers(),
                                                     self.diff_ids()):
      if this_digest == digest:
        return this_diff_id
    raise ValueError('Unmatched "digest": "%s"' % digest)

  def layer(self, diff_id):
    """Like `blob()`, but accepts the `diff_id` instead.

    The `diff_id` is the name for the digest of the uncompressed layer.

    Args:
      diff_id: the 'algo:digest' of the layer being addressed.

    Returns:
      The raw compressed blob bytes of the layer.
    """
    return self.blob(self._diff_id_to_digest(diff_id))

  def uncompressed_layer(self, diff_id):
    """Same as layer() but uncompressed."""
    return self.uncompressed_blob(self._diff_id_to_digest(diff_id))

  # __enter__ and __exit__ allow use as a context manager.
  @abc.abstractmethod
  def __enter__(self):
    """Open the image for reading."""

  @abc.abstractmethod
  def __exit__(self, unused_type, unused_value, unused_traceback):
    """Close the image."""

  def __str__(self):
    """A human-readable representation of the image."""
    return str(type(self))


class Delegate(DockerImage):
  """Forwards calls to the underlying image."""

  def __init__(self, image):
    """Constructor.

    Args:
      image: a DockerImage on which __enter__ has already been called.
    """
    self._image = image

  def manifest(self):
    """Override."""
    return self._image.manifest()

  def media_type(self):
    """Override."""
    return self._image.media_type()

  def diff_ids(self):
    """Override."""
    return self._image.diff_ids()

  def fs_layers(self):
    """Override."""
    return self._image.fs_layers()

  def config_blob(self):
    """Override."""
    return self._image.config_blob()

  def blob_set(self):
    """Override."""
    return self._image.blob_set()

  def config_file(self):
    """Override."""
    return self._image.config_file()

  def blob_size(self, digest):
    """Override."""
    return self._image.blob_size(digest)

  def blob(self, digest):
    """Override."""
    return self._image.blob(digest)

  def uncompressed_blob(self, digest):
    """Override."""
    return self._image.uncompressed_blob(digest)

  def layer(self, diff_id):
    """Override."""
    return self._image.layer(diff_id)

  def uncompressed_layer(self, diff_id):
    """Override."""
    return self._image.uncompressed_layer(diff_id)

  def __str__(self):
    """Override."""
    return str(self._image)


class FromRegistry(DockerImage):
  """This accesses a docker image hosted on a registry (non-local)."""

  def __init__(self,
               name,
               basic_creds,
               transport,
               accepted_mimes = docker_http.MANIFEST_SCHEMA2_MIMES):
    self._name = name
    self._creds = basic_creds
    self._original_transport = transport
    self._accepted_mimes = accepted_mimes
    self._response = {}

  def _content(self,
               suffix,
               accepted_mimes = None,
               cache = True):
    """Fetches content of the resources from registry by http calls."""
    if isinstance(self._name, docker_name.Repository):
      suffix = '{repository}/{suffix}'.format(
          repository=self._name.repository, suffix=suffix)

    if suffix in self._response:
      return self._response[suffix]

    _, content = self._transport.Request(
        '{scheme}://{registry}/v2/{suffix}'.format(
            scheme=docker_http.Scheme(self._name.registry),
            registry=self._name.registry,
            suffix=suffix),
        accepted_codes=[six.moves.http_client.OK],
        accepted_mimes=accepted_mimes)
    if cache:
      self._response[suffix] = content
    return content

  def _tags(self):
    # See //cloud/containers/registry/proto/v2/tags.proto
    # for the full response structure.
    return json.loads(self._content('tags/list').decode('utf8'))

  def tags(self):
    return self._tags().get('tags', [])

  def manifests(self):
    payload = self._tags()
    if 'manifest' not in payload:
      # Only GCR supports this schema.
      return {}
    return payload['manifest']

  def children(self):
    payload = self._tags()
    if 'child' not in payload:
      # Only GCR supports this schema.
      return []
    return payload['child']

  def exists(self):
    try:
      manifest = json.loads(self.manifest(validate=False))
      return (manifest['schemaVersion'] == 2 and 'layers' in manifest and
              self.media_type() in self._accepted_mimes)
    except docker_http.V2DiagnosticException as err:
      if err.status == six.moves.http_client.NOT_FOUND:
        return False
      raise

  def manifest(self, validate=True):
    """Override."""
    # GET server1/v2/<name>/manifests/<tag_or_digest>

    if isinstance(self._name, docker_name.Tag):
      path = 'manifests/' + self._name.tag
      return self._content(path, self._accepted_mimes).decode('utf8')
    else:
      assert isinstance(self._name, docker_name.Digest)
      c = self._content('manifests/' + self._name.digest, self._accepted_mimes)
      computed = docker_digest.SHA256(c)
      if validate and computed != self._name.digest:
        raise DigestMismatchedError(
            'The returned manifest\'s digest did not match requested digest, '
            '%s vs. %s' % (self._name.digest, computed))
      return c.decode('utf8')

  def config_file(self):
    """Override."""
    return self.blob(self.config_blob()).decode('utf8')

  def blob_size(self, digest):
    """The byte size of the raw blob."""
    suffix = 'blobs/' + digest
    if isinstance(self._name, docker_name.Repository):
      suffix = '{repository}/{suffix}'.format(
          repository=self._name.repository, suffix=suffix)

    resp, unused_content = self._transport.Request(
        '{scheme}://{registry}/v2/{suffix}'.format(
            scheme=docker_http.Scheme(self._name.registry),
            registry=self._name.registry,
            suffix=suffix),
        method='HEAD',
        accepted_codes=[six.moves.http_client.OK])

    return int(resp['content-length'])

  # Large, do not memoize.
  def blob(self, digest):
    """Override."""
    # GET server1/v2/<name>/blobs/<digest>
    c = self._content('blobs/' + digest, cache=False)
    computed = docker_digest.SHA256(c)
    if digest != computed:
      raise DigestMismatchedError(
          'The returned content\'s digest did not match its content-address, '
          '%s vs. %s' % (digest, computed if c else '(content was empty)'))
    return c

  def catalog(self, page_size = 100):
    # TODO(user): Handle docker_name.Repository for /v2/<name>/_catalog
    if isinstance(self._name, docker_name.Repository):
      raise ValueError('Expected docker_name.Registry for "name"')

    url = '{scheme}://{registry}/v2/_catalog?n={page_size}'.format(
        scheme=docker_http.Scheme(self._name.registry),
        registry=self._name.registry,
        page_size=page_size)

    for _, content in self._transport.PaginatedRequest(
        url, accepted_codes=[six.moves.http_client.OK]):
      wrapper_object = json.loads(content.decode('utf8'))

      if 'repositories' not in wrapper_object:
        raise docker_http.BadStateException(
            'Malformed JSON response: %s' % content)

      # TODO(user): This should return docker_name.Repository
      for repo in wrapper_object['repositories']:
        yield repo

  # __enter__ and __exit__ allow use as a context manager.
  def __enter__(self):
    # Create a v2 transport to use for making authenticated requests.
    self._transport = docker_http.Transport(
        self._name, self._creds, self._original_transport, docker_http.PULL)

    return self

  def __exit__(self, unused_type, unused_value, unused_traceback):
    pass

  def __str__(self):
    return '<docker_image.FromRegistry name: {}>'.format(str(self._name))


# Gzip injects a timestamp into its output, which makes its output and digest
# non-deterministic.  To get reproducible pushes, freeze time.
# This approach is based on the following StackOverflow answer:
# http://stackoverflow.com/questions/264224/setting-the-gzip-timestamp-from-python
class _FakeTime(object):

  def time(self):
    return 1225856967.109


gzip.time = _FakeTime()


# Checks the contents of a file for magic bytes that indicate that it's gzipped
def is_compressed(name):
  return name[0:2] == b'\x1f\x8b'


class FromTarball(DockerImage):
  """This decodes the image tarball output of docker_build for upload."""

  def __init__(
      self,
      tarball,
      name = None,
      compresslevel = 9,
  ):
    self._tarball = tarball
    self._compresslevel = compresslevel
    self._memoize = {}
    self._lock = threading.Lock()
    self._name = name
    self._manifest = None
    self._blob_names = None
    self._config_blob = None

  # Layers can come in two forms, as an uncompressed tar in a directory
  # or as a gzipped tar. We need to account for both options, and be able
  # to return both uncompressed and compressed data.
  def _content(self,
               name,
               memoize = True,
               should_be_compressed = False):
    """Fetches a particular path's contents from the tarball."""
    # Check our cache
    if memoize:
      with self._lock:
        if (name, should_be_compressed) in self._memoize:
          return self._memoize[(name, should_be_compressed)]

    # tarfile is inherently single-threaded:
    # https://mail.python.org/pipermail/python-bugs-list/2015-March/265999.html
    # so instead of locking, just open the tarfile for each file
    # we want to read.
    with tarfile.open(name=self._tarball, mode='r') as tar:
      try:
        # If the layer is compressed and we need to return compressed
        # or if it's uncompressed and we need to return uncompressed
        # then return the contents as is.
        f = tar.extractfile(str(name))
        content = f.read()  # pytype: disable=attribute-error
      except KeyError:
        content = tar.extractfile(
            str('./' + name)).read()  # pytype: disable=attribute-error
      # We need to compress before returning. Use gzip.
      if should_be_compressed and not is_compressed(content):
        buf = io.BytesIO()
        zipped = gzip.GzipFile(
            mode='wb', compresslevel=self._compresslevel, fileobj=buf)
        try:
          zipped.write(content)
        finally:
          zipped.close()
        content = buf.getvalue()
      # The layer is gzipped but we need to return the uncompressed content
      # Open up the gzip and read the contents after.
      elif not should_be_compressed and is_compressed(content):
        buf = io.BytesIO(content)
        raw = gzip.GzipFile(mode='rb', fileobj=buf)
        content = raw.read()
      # Populate our cache.
      if memoize:
        with self._lock:
          self._memoize[(name, should_be_compressed)] = content
      return content

  def _gzipped_content(self, name):
    """Returns the result of _content with gzip applied."""
    return self._content(name, memoize=False, should_be_compressed=True)

  def _populate_manifest_and_blobs(self):
    """Populates self._manifest and self._blob_names."""
    config_blob = docker_digest.SHA256(self.config_file().encode('utf8'))
    manifest = {
        'mediaType': docker_http.MANIFEST_SCHEMA2_MIME,
        'schemaVersion': 2,
        'config': {
            'digest': config_blob,
            'mediaType': docker_http.CONFIG_JSON_MIME,
            'size': len(self.config_file())
        },
        'layers': [
            # Populated below
        ]
    }

    blob_names = {}

    config = json.loads(self.config_file())
    diff_ids = config['rootfs']['diff_ids']

    for i, layer in enumerate(self._layers):
      name = None
      diff_id = diff_ids[i]
      media_type = docker_http.LAYER_MIME
      size = 0
      urls = []

      if diff_id in self._layer_sources:
        # _layer_sources contains foreign layers from the base image
        name = self._layer_sources[diff_id]['digest']
        media_type = self._layer_sources[diff_id]['mediaType']
        size = self._layer_sources[diff_id]['size']
        if 'urls' in self._layer_sources[diff_id]:
          urls = self._layer_sources[diff_id]['urls']
      else:
        content = self._gzipped_content(layer)
        name = docker_digest.SHA256(content)
        size = len(content)

      blob_names[name] = layer

      layer_manifest = {
          'digest': name,
          'mediaType': media_type,
          'size': size,
      }

      if urls:
        layer_manifest['urls'] = urls

      manifest['layers'].append(layer_manifest)

    with self._lock:
      self._manifest = manifest
      self._blob_names = blob_names
      self._config_blob = config_blob

  def manifest(self):
    """Override."""
    if not self._manifest:
      self._populate_manifest_and_blobs()
    return json.dumps(self._manifest, sort_keys=True)

  def config_file(self):
    """Override."""
    return self._content(self._config_file).decode('utf8')

  # Could be large, do not memoize
  def uncompressed_blob(self, digest):
    """Override."""
    if not self._blob_names:
      self._populate_manifest_and_blobs()
    return self._content(
        self._blob_names[digest],
        memoize=False,
        should_be_compressed=False)

  # Could be large, do not memoize
  def blob(self, digest):
    """Override."""
    if not self._blob_names:
      self._populate_manifest_and_blobs()
    if digest == self._config_blob:
      return self.config_file().encode('utf8')
    return self._gzipped_content(
        self._blob_names[digest])

  # Could be large, do not memoize
  def uncompressed_layer(self, diff_id):
    """Override."""
    for (layer, this_diff_id) in zip(reversed(self._layers), self.diff_ids()):
      if diff_id == this_diff_id:
        return self._content(layer, memoize=False, should_be_compressed=False)
    raise ValueError('Unmatched "diff_id": "%s"' % diff_id)

  def _resolve_tag(self):
    """Resolve the singleton tag this tarball contains using legacy methods."""
    repo_bytes = self._content('repositories', memoize=False)
    repositories = json.loads(repo_bytes.decode('utf8'))
    if len(repositories) != 1:
      raise ValueError('Tarball must contain a single repository, '
                       'or a name must be specified to FromTarball.')

    for (repo, tags) in six.iteritems(repositories):
      if len(tags) != 1:
        raise ValueError('Tarball must contain a single tag, '
                         'or a name must be specified to FromTarball.')
      for (tag, unused_layer) in six.iteritems(tags):
        return '{repository}:{tag}'.format(repository=repo, tag=tag)

    raise Exception('unreachable')

  # __enter__ and __exit__ allow use as a context manager.
  def __enter__(self):
    manifest_json = self._content('manifest.json').decode('utf8')
    manifest_list = json.loads(manifest_json)

    config = None
    layers = []
    layer_sources = []
    # Find the right entry, either:
    # 1) We were supplied with an image name, which we must find in an entry's
    #   RepoTags, or
    # 2) We were not supplied with an image name, and this must have a single
    #   image defined.
    if len(manifest_list) != 1:
      if not self._name:
        # If we run into this situation, fall back on the legacy repositories
        # file to tell us the single tag.  We do this because Bazel will apply
        # build targets as labels, so each layer will be labelled, but only
        # the final label will appear in the resulting repositories file.
        self._name = self._resolve_tag()

    for entry in manifest_list:
      if not self._name or str(self._name) in (entry.get('RepoTags') or []):
        config = entry.get('Config')
        layers = entry.get('Layers', [])
        layer_sources = entry.get('LayerSources', {})

    if not config:
      raise ValueError('Unable to find %s in provided tarball.' % self._name)

    # Metadata from the tarball's configuration we need to construct the image.
    self._config_file = config
    self._layers = layers
    self._layer_sources = layer_sources

    # We populate "manifest" and "blobs" lazily for two reasons:
    # 1) Allow use of this library for reading the config_file() from the image
    #   layer shards Bazel produces.
    # 2) Performance of the case where all we read is the config_file().

    return self

  def __exit__(self, unused_type, unused_value, unused_traceback):
    pass


class FromDisk(DockerImage):
  """This accesses a more efficient on-disk format than FromTarball.

  FromDisk reads an on-disk format optimized for use with push and pull.

  It is expected that the number of layers in config_file's rootfs.diff_ids
  matches: count(legacy_base.layers) + len(layers).

  Layers are drawn from legacy_base first (it is expected to be the base),
  and then from layers.

  This is effectively the dual of the save.fast method, and is intended for use
  with Bazel's rules_docker.

  Args:
    config_file: the contents of the config file.
    layers: a list of pairs.  The first element is the path to a file containing
        the second element's sha256.  The second element is the .tar.gz of a
        filesystem layer.  These are ordered as they'd appear in the manifest.
    uncompressed_layers: Optionally, a list of pairs. The first element is the
        path to a file containing the second element's sha256.
        The second element is the .tar of a filesystem layer.
    legacy_base: Optionally, the path to a legacy base image in FromTarball form
    foreign_layers_manifest: Optionally a tar manifest from the base
        image that describes the ForeignLayers needed by this image.
  """

  def __init__(self,
               config_file,
               layers,
               uncompressed_layers = None,
               legacy_base = None,
               foreign_layers_manifest = None):
    self._config = config_file
    self._manifest = None
    self._foreign_layers_manifest = foreign_layers_manifest
    self._layers = []
    self._layer_to_filename = {}
    for (name_file, content_file) in layers:
      with io.open(name_file, u'r') as reader:
        layer_name = 'sha256:' + reader.read()
      self._layers.append(layer_name)
      self._layer_to_filename[layer_name] = content_file

    self._uncompressed_layers = []
    self._uncompressed_layer_to_filename = {}
    if uncompressed_layers:
      for (name_file, content_file) in uncompressed_layers:
        with io.open(name_file, u'r') as reader:
          layer_name = 'sha256:' + reader.read()
        self._uncompressed_layers.append(layer_name)
        self._uncompressed_layer_to_filename[layer_name] = content_file

    self._legacy_base = None
    if legacy_base:
      with FromTarball(legacy_base) as base:
        self._legacy_base = base

  def _get_foreign_layers(self):
    foreign_layers = []
    if self._foreign_layers_manifest:
      manifest = json.loads(self._foreign_layers_manifest)
      if 'layers' in manifest:
        for layer in manifest['layers']:
          if layer['mediaType'] == docker_http.FOREIGN_LAYER_MIME:
            foreign_layers.append(layer)
    return foreign_layers

  def _get_foreign_layer_by_digest(self, digest):
    for foreign_layer in self._get_foreign_layers():
      if foreign_layer['digest'] == digest:
        return foreign_layer
    return None

  def _populate_manifest(self):
    base_layers = []
    if self._legacy_base:
      base_layers = json.loads(self._legacy_base.manifest())['layers']
    elif self._foreign_layers_manifest:
      # Manifest files found in tar files are actually a json list.
      # This code iterates through that collection and appends any foreign
      # layers described in the order found in the config file.
      base_layers += self._get_foreign_layers()

    # TODO(user): Update mimes here for oci_compat.
    self._manifest = json.dumps(
        {
            'schemaVersion':
                2,
            'mediaType':
                docker_http.MANIFEST_SCHEMA2_MIME,
            'config': {
                'mediaType':
                    docker_http.CONFIG_JSON_MIME,
                'size':
                    len(self.config_file()),
                'digest':
                    docker_digest.SHA256(self.config_file().encode('utf8'))
            },
            'layers':
                base_layers + [{
                    'mediaType': docker_http.LAYER_MIME,
                    'size': self.blob_size(digest),
                    'digest': digest
                } for digest in self._layers]
        },
        sort_keys=True)

  def manifest(self):
    """Override."""
    if not self._manifest:
      self._populate_manifest()
    return self._manifest

  def config_file(self):
    """Override."""
    return self._config

  # Could be large, do not memoize
  def uncompressed_blob(self, digest):
    """Override."""
    if digest not in self._layer_to_filename:
      if self._get_foreign_layer_by_digest(digest):
        return bytes([])
      else:
        # Leverage the FromTarball fast-path.
        return self._legacy_base.uncompressed_blob(digest)
    return super(FromDisk, self).uncompressed_blob(digest)

  def uncompressed_layer(self, diff_id):
    if diff_id in self._uncompressed_layer_to_filename:
      with io.open(self._uncompressed_layer_to_filename[diff_id],
                   u'rb') as reader:
        # TODO(b/118349036): Remove the disable once the pytype bug is fixed.
        return reader.read()  # pytype: disable=bad-return-type
    if self._legacy_base and diff_id in self._legacy_base.diff_ids():
      return self._legacy_base.uncompressed_layer(diff_id)
    return super(FromDisk, self).uncompressed_layer(diff_id)

  # Could be large, do not memoize
  def blob(self, digest):
    """Override."""
    if digest not in self._layer_to_filename:
      return self._legacy_base.blob(digest)
    with open(self._layer_to_filename[digest], 'rb') as reader:
      return reader.read()

  def blob_size(self, digest):
    """Override."""
    if digest not in self._layer_to_filename:
      return self._legacy_base.blob_size(digest)
    info = os.stat(self._layer_to_filename[digest])
    return info.st_size

  # __enter__ and __exit__ allow use as a context manager.
  def __enter__(self):
    return self

  def __exit__(self, unused_type, unused_value, unused_traceback):
    pass


def _in_whiteout_dir(fs, name):
  while name:
    dirname = os.path.dirname(name)
    if name == dirname:
      break
    if fs.get(dirname):
      return True
    name = dirname
  return False


_WHITEOUT_PREFIX = '.wh.'


def extract(image, tar):
  """Extract the final filesystem from the image into tar.

  Args:
    image: a docker image whose final filesystem to construct.
    tar: the tarfile into which we are writing the final filesystem.
  """
  # Maps all of the files we have already added (and should never add again)
  # to whether they are a tombstone or not.
  fs = {}

  # Walk the layers, topmost first and add files.  If we've seen them in a
  # higher layer then we skip them
  for layer in image.diff_ids():
    buf = io.BytesIO(image.uncompressed_layer(layer))
    with tarfile.open(mode='r:', fileobj=buf) as layer_tar:
      for tarinfo in layer_tar:
        # If we see a whiteout file, then don't add anything to the tarball
        # but ensure that any lower layers don't add a file with the whited
        # out name.
        basename = os.path.basename(tarinfo.name)
        dirname = os.path.dirname(tarinfo.name)
        tombstone = basename.startswith(_WHITEOUT_PREFIX)
        if tombstone:
          basename = basename[len(_WHITEOUT_PREFIX):]

        # Before adding a file, check to see whether it (or its whiteout) have
        # been seen before.
        name = os.path.normpath(os.path.join('.', dirname, basename))
        if name in fs:
          continue

        # Check for a whited out parent directory
        if _in_whiteout_dir(fs, name):
          continue

        # Mark this file as handled by adding its name.
        # A non-directory implicitly tombstones any entries with
        # a matching (or child) name.
        fs[name] = tombstone or not tarinfo.isdir()
        if not tombstone:
          if tarinfo.isfile():
            tar.addfile(tarinfo, fileobj=layer_tar.extractfile(tarinfo))
          else:
            tar.addfile(tarinfo, fileobj=None)



import json
import logging

from docker import APIClient

from kubeflow.fairing.builders.base_builder import BaseBuilder
from kubeflow.fairing.builders import dockerfile
from kubeflow.fairing.constants import constants

logger = logging.getLogger(__name__)


class DockerBuilder(BaseBuilder):
    """A builder using the local Docker client"""

    def __init__(self,
                 registry=None,
                 image_name=constants.DEFAULT_IMAGE_NAME,
                 base_image=constants.DEFAULT_BASE_IMAGE,
                 preprocessor=None,
                 push=True,
                 dockerfile_path=None):
        super().__init__(
            registry=registry,
            image_name=image_name,
            push=push,
            base_image=base_image,
            preprocessor=preprocessor,
            dockerfile_path=dockerfile_path)

    def build(self):
        logging.info("Building image using docker")
        self.docker_client = APIClient(version='auto')
        self._build()
        if self.push:
            self.publish()

    def _build(self):
        """build the docker image"""
        docker_command = self.preprocessor.get_command()
        logger.warning("Docker command: {}".format(docker_command))
        if not docker_command:
            logger.warning("Not setting a command for the output docker image.")
        install_reqs_before_copy = self.preprocessor.is_requirements_txt_file_present()
        if self.dockerfile_path:
            dockerfile_path = self.dockerfile_path
        else:
            dockerfile_path = dockerfile.write_dockerfile(
                docker_command=docker_command,
                path_prefix=self.preprocessor.path_prefix,
                base_image=self.base_image,
                install_reqs_before_copy=install_reqs_before_copy)
        self.preprocessor.output_map[dockerfile_path] = 'Dockerfile'
        context_file, context_hash = self.preprocessor.context_tar_gz()
        self.image_tag = self.full_image_name(context_hash)
        logger.warning('Building docker image {}...'.format(self.image_tag))
        with open(context_file, 'rb') as fileobj:
            bld = self.docker_client.build(
                path='.',
                custom_context=True,
                fileobj=fileobj,
                tag=self.image_tag,
                encoding='utf-8'
            )
        for line in bld:
            self._process_stream(line)

    def publish(self):
        """push the docker image to the docker registry"""
        logger.warning('Publishing image {}...'.format(self.image_tag))
        for line in self.docker_client.push(self.image_tag, stream=True):
            self._process_stream(line)

    def _process_stream(self, line):
        """
        Parse the docker command output by line

        :param line:  a line of the command output message

        """
        raw = line.decode('utf-8').strip()
        lns = raw.split('\n')
        for ln in lns:
            try:
                ljson = json.loads(ln)
                if ljson.get('error'):
                    msg = str(ljson.get('error', ljson))
                    logger.error('Build failed: %s', msg)
                    raise Exception('Image build failed: ' + msg)
                else:
                    if ljson.get('stream'):
                        msg = 'Build output: {}'.format(
                            ljson['stream'].strip())
                    elif ljson.get('status'):
                        msg = 'Push output: {} {}'.format(
                            ljson['status'],
                            ljson.get('progress')
                        )
                    elif ljson.get('aux'):
                        msg = 'Push finished: {}'.format(ljson.get('aux'))
                    else:
                        msg = str(ljson)
                    logger.info(msg)

            except json.JSONDecodeError:
                logger.warning('JSON decode error: {}'.format(ln))



import google.auth
from google.cloud import storage
from google.cloud.exceptions import NotFound
from kubeflow.fairing.constants import constants
from kubernetes import client
import logging

import os
import json

logger = logging.getLogger(__name__)


class GCSUploader(object):
    def __init__(
            self,
            credentials_file=os.environ.get(constants.GOOGLE_CREDS_ENV)): #pylint:disable=unused-argument
        self.storage_client = storage.Client()

    def upload_to_bucket(self,
                         blob_name,
                         bucket_name,
                         file_to_upload):
        bucket = self.get_or_create_bucket(bucket_name)
        blob = bucket.blob(blob_name)
        blob.upload_from_filename(file_to_upload)
        return "gs://{}/{}".format(bucket_name, blob_name)

    def get_or_create_bucket(self, bucket_name):
        try:
            bucket = self.storage_client.get_bucket(bucket_name)
            return bucket
        except NotFound:
            bucket = self.storage_client.create_bucket(bucket_name)
            return bucket


def guess_project_name(credentials_file=None):
    # Check credentials file if provided
    if credentials_file is not None:
        with open(credentials_file, 'r') as f:
            data = f.read()
        parsed_creds_file = json.loads(data)
        return parsed_creds_file['project_id']

    # Use the google.auth library to retrieve credentials. Checks the following
    # locations in order:
    # 1. GOOGLE_APPLICATION_CREDENTIALS environment variable
    # 2. Google Cloud SDK (gcloud)
    # 3. App Engine Identity service
    # 4. Compute Engine Metadata service
    credentials, project_id = google.auth.default() #pylint:disable=unused-variable

    if project_id is None:
        raise Exception('Could not determine project id.')

    return project_id


def add_gcp_credentials_if_exists(kube_manager, pod_spec, namespace):
    try:
        if kube_manager.secret_exists(constants.GCP_CREDS_SECRET_NAME, namespace):
            add_gcp_credentials(kube_manager, pod_spec, namespace)
        else:
            logger.warning("Not able to find gcp credentials secret: {}"\
                           .format(constants.GCP_CREDS_SECRET_NAME))
            logger.warning("Trying workload identity service account: {}"\
                           .format(constants.GCP_SERVICE_ACCOUNT_NAME))
            pod_spec.service_account_name = constants.GCP_SERVICE_ACCOUNT_NAME
    except Exception as e:
        logger.warning("could not check for secret: {}".format(e))


def add_gcp_credentials(kube_manager, pod_spec, namespace):
    """Note: This method will be deprecated soon and will become unavailable by 1.0.
             All future access will use Workload Identity.
    """
    if not kube_manager.secret_exists(constants.GCP_CREDS_SECRET_NAME, namespace):
        raise ValueError('Unable to mount credentials: '
                         + 'Secret user-gcp-sa not found in namespace {}'.format(namespace))

    # Set appropriate secrets and volumes to enable kubeflow-user service
    # account.
    env_var = client.V1EnvVar(
        name='GOOGLE_APPLICATION_CREDENTIALS',
        value='/etc/secrets/user-gcp-sa.json')
    if pod_spec.containers[0].env:
        pod_spec.containers[0].env.append(env_var)
    else:
        pod_spec.containers[0].env = [env_var]

    volume_mount = client.V1VolumeMount(
        name='user-gcp-sa', mount_path='/etc/secrets', read_only=True)
    if pod_spec.containers[0].volume_mounts:
        pod_spec.containers[0].volume_mounts.append(volume_mount)
    else:
        pod_spec.containers[0].volume_mounts = [volume_mount]

    volume = client.V1Volume(
        name='user-gcp-sa',
        secret=client.V1SecretVolumeSource(secret_name='user-gcp-sa'))
    if pod_spec.volumes:
        pod_spec.volumes.append(volume)
    else:
        pod_spec.volumes = [volume]

def get_default_docker_registry():
    try:
        return 'gcr.io/{}/fairing-job'.format(guess_project_name())
    except Exception:
        return None



import tempfile
import os
import posixpath
import stat
import logging
import collections

from kubeflow.fairing import utils as fairing_utils
from kubeflow.fairing.preprocessors.base import BasePreProcessor
from kubeflow.fairing.builders.append.append import AppendBuilder
from kubeflow.fairing.deployers.job.job import Job
from kubeflow.fairing.deployers.tfjob.tfjob import TfJob
from kubeflow.fairing.constants import constants
from kubeflow.fairing.kubernetes import utils as k8s_utils
from kubeflow.fairing.cloud import storage
from kubeflow.fairing.cloud import gcp
from kubeflow.fairing.frameworks import lightgbm_dist_training_init
from kubeflow.fairing.frameworks import utils

logger = logging.getLogger(__name__)

TRAIN_DATA_FIELDS = ["data", "train", "train_data",
                     "train_data_file", "data_filename"]
TEST_DATA_FIELDS = ["valid", "test", "valid_data", "valid_data_file", "test_data",
                    "test_data_file", "valid_filenames"]
NUM_MACHINES_FILEDS = ["num_machines", "num_machine"]
PORT_FIELDS = ["local_listen_port", "local_port"]
MLIST_FIELDS = ["machine_list_filename",
                "machine_list_file", "machine_list", "mlist"]
OUTPUT_MODEL_FIELDS = ["output_model", "model_output", "model_out"]
INPUT_MODEL_FIELDS = ["input_model", "model_input", "model_in"]
OUTPUT_RESULT_FIELDS = ["output_result", "predict_result", "prediction_result",
                        "predict_name", "prediction_name", "pred_name", "name_pred"]
MACHINE_FIELDS = ["machines", "workers", "nodes"]
TREE_LEARNER_FIELDS = ["tree_learner",
                       "tree", "tree_type", "tree_learner_type"]
ENTRYPOINT = posixpath.join(constants.DEFAULT_DEST_PREFIX, "entrypoint.sh")
LIGHTGBM_EXECUTABLE = "lightgbm"
CONFIG_FILE_NAME = "config.conf"
MLIST_FILE_NAME = "mlist.txt"
BLACKLISTED_FIELDS = PORT_FIELDS + MLIST_FIELDS + MACHINE_FIELDS
WEIGHT_FILE_EXT = ".weight"
DATA_PARALLEL_MODES = ["data", "voting"]


def _modify_paths_in_config(config, field_names, dst_base_dir):
    """modify lightgbm config fields

    :param config: config entries
    :param field_names: list of fields
    :param dst_base_dir: path to destination directory

    """
    field_name, field_value = utils.get_config_value(config, field_names)
    if field_value is None:
        return [], []
    src_paths = field_value.split(",")
    dst_paths = []
    for src_path in src_paths:
        file_name = os.path.split(src_path)[-1]
        dst_paths.append(posixpath.join(dst_base_dir, file_name))
    config[field_name] = ",".join(dst_paths)
    return src_paths, dst_paths


def _update_maps(output_map, copy_files, src_paths, dst_paths):
    """update maps

    :param output_map: output map entries
    :param copy_files: files to be copied
    :param src_paths: source paths
    :param dst_paths: destination paths

    """
    for src_path, dst_path in zip(src_paths, dst_paths):
        if os.path.exists(src_path):
            output_map[src_path] = dst_path
        else:
            copy_files[src_path] = dst_path


def _get_commands_for_file_ransfer(files_map):
    """get commands for file transfer

    :param files_map: files to be mapped

    """
    cmds = []
    for k, v in files_map.items():
        storage_obj = storage.get_storage_class(k)()
        if storage_obj.exists(k):
            cmds.append(storage_obj.copy_cmd(k, v))
        else:
            raise RuntimeError("Remote file {} does't exist".format(k))
    return cmds


def _generate_entrypoint(copy_files_before, copy_files_after, config_file,
                         init_cmds=None, copy_patitioned_files=None):
    """ generate entry point

    :param copy_files_before: previous copied files
    :param copy_files_after: files to be copied
    :param config_file: path to config file
    :param init_cmds:  commands(Default value = None)
    :param copy_patitioned_files:  (Default value = None)

    """
    buf = ["#!/bin/sh",
           "set -e"]
    if init_cmds:
        buf.extend(init_cmds)
    # In data prallel mode, copying files based on RANK of the worker in the cluster.
    # The data is partitioned (#partitions=#workers) and each worker gets one partition of the data.
    if copy_patitioned_files and len(copy_patitioned_files) > 0: #pylint:disable=len-as-condition
        buf.append("case $RANK in")
        for rank, files in copy_patitioned_files.items():
            buf.append("\t{})".format(rank))
            buf.extend(
                ["\t\t" + cmd for cmd in _get_commands_for_file_ransfer(files)])
            buf.append("\t\t;;")
        buf.append("esac")

    # copying files that are common to all workers
    buf.extend(_get_commands_for_file_ransfer(copy_files_before))

    buf.append("echo 'All files are copied!'")
    buf.append("{} config={}".format(LIGHTGBM_EXECUTABLE, config_file))
    for k, v in copy_files_after.items():
        storage_obj = storage.get_storage_class(k)()
        buf.append(storage_obj.copy_cmd(v, k))
    _, file_name = tempfile.mkstemp()
    with open(file_name, 'w') as fh:
        content = "\n".join(buf)
        fh.write(content)
        fh.write("\n")
    st = os.stat(file_name)
    os.chmod(file_name, st.st_mode | stat.S_IEXEC)
    return file_name


def _add_train_weight_file(config, dst_base_dir):
    """add train weight file

    :param config: config entries
    :param dst_base_dir: destination directory

    """
    _, field_value = utils.get_config_value(config, TRAIN_DATA_FIELDS)
    if field_value is None:
        return [], []
    else:
        src_paths = field_value.split(",")
        weight_paths = [x+WEIGHT_FILE_EXT for x in src_paths]
        weight_paths_found = []
        weight_paths_dst = []
        for path in weight_paths:
            found = os.path.exists(path)
            if not found:
                # in case the path is local and doesn't exist
                storage_class = storage.lookup_storage_class(path)
                if storage_class:
                    found = storage_class().exists(path)
            if found:
                weight_paths_found.append(path)
                file_name = os.path.split(path)[-1]
                weight_paths_dst.append(
                    posixpath.join(dst_base_dir, file_name))
        return weight_paths_found, weight_paths_dst


def generate_context_files(config, config_file_name, num_machines):
    """generate context files

    :param config: config entries
    :param config_file_name: config file name
    :param num_machines: number of machines

    """
    # Using ordered dict to have consistent behaviour around order in which
    # files are copied in the worker nodes.
    output_map = collections.OrderedDict()
    copy_files_before = collections.OrderedDict()
    copy_files_after = collections.OrderedDict()
    copy_patitioned_files = collections.OrderedDict()

    # config will be modified inplace in this function so taking a copy
    config = config.copy()  # shallow copy is good enough

    _, tree_learner = utils.get_config_value(config, TREE_LEARNER_FIELDS)
    parition_data = tree_learner and tree_learner.lower() in DATA_PARALLEL_MODES
    remote_files = [(copy_files_before,
                     [TEST_DATA_FIELDS, INPUT_MODEL_FIELDS]),
                    (copy_files_after,
                     [OUTPUT_MODEL_FIELDS, OUTPUT_RESULT_FIELDS])]

    if parition_data:
        train_data_field, train_data_value = utils.get_config_value(
            config, TRAIN_DATA_FIELDS)
        train_files = train_data_value.split(",")
        if len(train_files) != num_machines:
            raise RuntimeError("#Training files listed in the {}={} field in the config should be "
                               "equal to the num_machines={} config value."\
                                .format(train_data_field, train_data_value, num_machines))
        weight_src_paths, weight_dst_paths = _add_train_weight_file(config,
                                                                    constants.DEFAULT_DEST_PREFIX)
        dst = posixpath.join(constants.DEFAULT_DEST_PREFIX, "train_data")
        config[train_data_field] = dst
        for i, f in enumerate(train_files):
            copy_patitioned_files[i] = collections.OrderedDict()
            copy_patitioned_files[i][f] = dst
            if f+WEIGHT_FILE_EXT in weight_src_paths:
                copy_patitioned_files[i][f +
                                         WEIGHT_FILE_EXT] = dst+WEIGHT_FILE_EXT
    else:
        train_data_field, train_data_value = utils.get_config_value(
            config, TRAIN_DATA_FIELDS)
        if len(train_data_value.split(",")) > 1:
            raise RuntimeError("{} has more than one file specified but tree-learner is set to {} "
                               "which can't handle multiple files. For distributing data across "
                               "multiple workers, please use one of {} as a tree-learner method. "
                               "For more information please refer the LightGBM parallel guide"
                               " https://github.com/microsoft/LightGBM/blob/master/docs/"
                               "Parallel-Learning-Guide.rst".format(
                                   train_data_field, tree_learner, DATA_PARALLEL_MODES))
        remote_files[0][1].insert(0, TRAIN_DATA_FIELDS)
        weight_src_paths, weight_dst_paths = _add_train_weight_file(config,
                                                                    constants.DEFAULT_DEST_PREFIX)
        _update_maps(output_map, copy_files_before, weight_src_paths, weight_dst_paths)

    for copy_files, field_names_list in remote_files:
        for field_names in field_names_list:
            src_paths, dst_paths = _modify_paths_in_config(
                config, field_names, constants.DEFAULT_DEST_PREFIX)
            _update_maps(output_map, copy_files, src_paths, dst_paths)

    if len(output_map) + len(copy_files_before) + len(copy_patitioned_files) == 0:
        raise RuntimeError("Both train and test data is missing in the config")
    modified_config_file_name = utils.save_properties_config_file(config)
    config_in_docker = posixpath.join(
        constants.DEFAULT_DEST_PREFIX, CONFIG_FILE_NAME)
    output_map[modified_config_file_name] = config_in_docker
    output_map[config_file_name] = config_in_docker + ".original"

    init_cmds = None
    if num_machines > 1:
        init_file = lightgbm_dist_training_init.__file__
        init_file_name = os.path.split(init_file)[1]
        output_map[init_file] = os.path.join(
            constants.DEFAULT_DEST_PREFIX, init_file_name)
        init_cmds = ["RANK=`python {} {} {}`".format(init_file_name,
                                                     CONFIG_FILE_NAME,
                                                     MLIST_FILE_NAME)]
    entrypoint_file_name = _generate_entrypoint(
        copy_files_before, copy_files_after, config_in_docker, init_cmds, copy_patitioned_files)
    output_map[entrypoint_file_name] = ENTRYPOINT
    output_map[utils.__file__] = os.path.join(
        constants.DEFAULT_DEST_PREFIX, "utils.py")

    return output_map


def execute(config,
            docker_registry,
            base_image="gcr.io/kubeflow-fairing/lightgbm:latest",
            namespace=None,
            stream_log=True,
            cores_per_worker=None,
            memory_per_worker=None,
            pod_spec_mutators=None):
    """Runs the LightGBM CLI in a single pod in user's Kubeflow cluster.
    Users can configure it to be a train, predict, and other supported tasks
    by using the right config.
    Please refere https://github.com/microsoft/LightGBM/blob/master/docs/Parameters.rst
    for more information on config options.

    :param config: config entries
    :param docker_registry: docker registry name
    :param base_image: base image (Default value = "gcr.io/kubeflow-fairing/lightgbm:latest")
    :param namespace: k8s namespace (Default value = None)
    :param stream_log: should that stream log? (Default value = True)
    :param cores_per_worker: number of cores per worker (Default value = None)
    :param memory_per_worker: memory value per worker (Default value = None)
    :param pod_spec_mutators: pod spec mutators (Default value = None)

    """
    if not namespace and not fairing_utils.is_running_in_k8s():
        namespace = "kubeflow"
    namespace = namespace or fairing_utils.get_default_target_namespace()
    config_file_name = None
    if isinstance(config, str):
        config_file_name = config
        config = utils.load_properties_config_file(config)
    elif isinstance(config, dict):
        config_file_name = utils.save_properties_config_file(config)
    else:
        raise RuntimeError("config should be of type dict or string(filepath) "
                           "but got {}".format(type(dict)))

    utils.scrub_fields(config, BLACKLISTED_FIELDS)

    _, num_machines = utils.get_config_value(config, NUM_MACHINES_FILEDS)
    num_machines = num_machines or 1
    if num_machines:
        try:
            num_machines = int(num_machines)
        except ValueError:
            raise ValueError("num_machines value in config should be an int >= 1 "
                             "but got {}".format(config.get('num_machines')))
        if num_machines < 1:
            raise ValueError(
                "num_machines value in config should >= 1 but got {}".format(num_machines))

    if num_machines > 1:
        config['machine_list_file'] = "mlist.txt"
    output_map = generate_context_files(
        config, config_file_name, num_machines)

    preprocessor = BasePreProcessor(
        command=[ENTRYPOINT], output_map=output_map)
    builder = AppendBuilder(registry=docker_registry,
                            base_image=base_image, preprocessor=preprocessor)
    builder.build()
    pod_spec = builder.generate_pod_spec()

    pod_spec_mutators = pod_spec_mutators or []
    pod_spec_mutators.append(gcp.add_gcp_credentials_if_exists)
    pod_spec_mutators.append(k8s_utils.get_resource_mutator(
        cores_per_worker, memory_per_worker))

    if num_machines == 1:
        # non-distributed mode
        deployer = Job(namespace=namespace,
                       pod_spec_mutators=pod_spec_mutators,
                       stream_log=stream_log)
    else:
        # distributed mode
        deployer = TfJob(namespace=namespace,
                         pod_spec_mutators=pod_spec_mutators,
                         chief_count=1,
                         worker_count=num_machines-1,
                         stream_log=stream_log)
    deployer.deploy(pod_spec)
    return deployer



# Copyright 2019 The Kubeflow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import logging

from kubernetes import client as k8s_client

from kfserving import V1alpha2EndpointSpec
from kfserving import V1alpha2PredictorSpec
from kfserving import V1alpha2TensorflowSpec
from kfserving import V1alpha2ONNXSpec
from kfserving import V1alpha2PyTorchSpec
from kfserving import V1alpha2SKLearnSpec
from kfserving import V1alpha2TensorRTSpec
from kfserving import V1alpha2XGBoostSpec
from kfserving import V1alpha2CustomSpec
from kfserving import V1alpha2InferenceServiceSpec
from kfserving import V1alpha2InferenceService

from kubeflow.fairing.constants import constants
from kubeflow.fairing.deployers.deployer import DeployerInterface
from kubeflow.fairing.kubernetes.manager import KubeManager
from kubeflow.fairing import utils

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class KFServing(DeployerInterface):
    """Serves a prediction endpoint using Kubeflow KFServing."""

    def __init__(self, framework, default_storage_uri=None, canary_storage_uri=None,
                 canary_traffic_percent=0, namespace=None, labels=None, annotations=None,
                 custom_default_container=None, custom_canary_container=None,
                 stream_log=False, cleanup=False):
        """
        :param framework: The framework for the InferenceService, such as Tensorflow,
            XGBoost and ScikitLearn etc.
        :param default_storage_uri: URI pointing to Saved Model assets for default service.
        :param canary_storage_uri: URI pointing to Saved Model assets for canary service.
        :param canary_traffic_percent: The amount of traffic to sent to the canary, defaults to 0.
        :param namespace: The k8s namespace where the InferenceService will be deployed.
        :param labels: Labels for the InferenceService, separate with commas if have more than one.
        :param annotations: Annotations for the InferenceService,
            separate with commas if have more than one.
        :param custom_default_container: A flexible custom default container for arbitrary customer
                                 provided containers.
        :param custom_canary_container: A flexible custom canary container for arbitrary customer
                                 provided containers.
        :param stream_log: Show log or not when InferenceService started, defaults to True.
        :param cleanup: Delete the kfserving or not, defaults to False.
        """
        self.framework = framework
        self.default_storage_uri = default_storage_uri
        self.canary_storage_uri = canary_storage_uri
        self.canary_traffic_percent = canary_traffic_percent
        self.annotations = annotations
        self.set_labels(labels)
        self.cleanup = cleanup
        self.custom_default_container = custom_default_container
        self.custom_canary_container = custom_canary_container
        self.stream_log = stream_log
        self.backend = KubeManager()

        if namespace is None:
            self.namespace = utils.get_default_target_namespace()
        else:
            self.namespace = namespace

        if self.framework != 'custom' and self.default_storage_uri is None:
            raise RuntimeError("The default_storage_uri must be specified for "
                               "{} framework.".format(self.framework))
        if self.framework == 'custom' and self.custom_default_container is None:
            raise RuntimeError("The custom_default_container must be specified "
                               "for custom framework.")

    def set_labels(self, labels):
        """set label for deployed prediction

        :param labels: dictionary of labels {label_name:label_value}

        """
        self.fairing_id = str(uuid.uuid1())
        self.labels = {'fairing-id': self.fairing_id}
        if labels:
            self.labels.update(labels)

    def deploy(self, isvc):  # pylint:disable=arguments-differ,unused-argument
        """deploy kfserving endpoint

        :param isvc: InferenceService for deploying.

        """
        self.created_isvc = self.backend.create_isvc(
            self.namespace, self.generate_isvc())

        if self.stream_log:
            self.get_logs()

        isvc_name = self.created_isvc['metadata']['name']
        logger.info(
            "Deployed the InferenceService {} successfully.".format(isvc_name))

        if self.cleanup:
            logger.warning(
                "Cleaning up InferenceService {}...".format(isvc_name))
            self.backend.delete_isvc(isvc_name, self.namespace)

        return isvc_name

    def generate_isvc(self):
        """ generate InferenceService """

        api_version = constants.KFSERVING_GROUP + '/' + constants.KFSERVING_VERSION
        default_predictor, canary_predictor = None, None

        if self.framework == 'custom':
            default_predictor = self.generate_predictor_spec(
                self.framework, container=self.custom_default_container)
        else:
            default_predictor = self.generate_predictor_spec(
                self.framework, storage_uri=self.default_storage_uri)

        if self.framework != 'custom' and self.canary_storage_uri is not None:
            canary_predictor = self.generate_predictor_spec(
                self.framework, storage_uri=self.canary_storage_uri)
        if self.framework == 'custom' and self.custom_canary_container is not None:
            canary_predictor = self.generate_predictor_spec(
                self.framework, container=self.custom_canary_container)

        if canary_predictor:
            isvc_spec = V1alpha2InferenceServiceSpec(
                default=V1alpha2EndpointSpec(predictor=default_predictor),
                canary=V1alpha2EndpointSpec(predictor=canary_predictor),
                canary_traffic_percent=self.canary_traffic_percent)
        else:
            isvc_spec = V1alpha2InferenceServiceSpec(
                default=V1alpha2EndpointSpec(predictor=default_predictor),
                canary_traffic_percent=self.canary_traffic_percent)

        return V1alpha2InferenceService(api_version=api_version,
                                        kind=constants.KFSERVING_KIND,
                                        metadata=k8s_client.V1ObjectMeta(
                                            generate_name=constants.KFSERVING_DEFAULT_NAME,
                                            namespace=self.namespace),
                                        spec=isvc_spec)


    def generate_predictor_spec(self, framework, storage_uri=None, container=None):
        '''Generate predictor spec according to framework and
           default_storage_uri or custom container.
        '''
        if self.framework == 'tensorflow':
            predictor = V1alpha2PredictorSpec(
                tensorflow=V1alpha2TensorflowSpec(storage_uri=storage_uri))
        elif self.framework == 'onnx':
            predictor = V1alpha2PredictorSpec(
                onnx=V1alpha2ONNXSpec(storage_uri=storage_uri))
        elif self.framework == 'pytorch':
            predictor = V1alpha2PredictorSpec(
                pytorch=V1alpha2PyTorchSpec(storage_uri=storage_uri))
        elif self.framework == 'sklearn':
            predictor = V1alpha2PredictorSpec(
                sklearn=V1alpha2SKLearnSpec(storage_uri=storage_uri))
        elif self.framework == 'tensorrt':
            predictor = V1alpha2PredictorSpec(
                tensorrt=V1alpha2TensorRTSpec(storage_uri=storage_uri))
        elif self.framework == 'xgboost':
            predictor = V1alpha2PredictorSpec(
                xgboost=V1alpha2XGBoostSpec(storage_uri=V1alpha2XGBoostSpec))
        elif self.framework == 'custom':
            predictor = V1alpha2PredictorSpec(
                custom=V1alpha2CustomSpec(container=container))
        else:
            raise RuntimeError("Unsupported framework {}".format(framework))
        return predictor

    def get_logs(self):
        """ get log from prediction pod"""
        name = self.created_isvc['metadata']['name']
        namespace = self.created_isvc['metadata']['namespace']

        self.backend.log(name, namespace, self.labels,
                         container=constants.KFSERVING_CONTAINER_NAME, follow=False)



import setuptools

with open('requirements.txt') as f:
    REQUIRES = f.read().splitlines()

setuptools.setup(
    name='kubeflow-fairing',
    version='0.7.0.1',
    author="Kubeflow Authors",
    author_email='hejinchi@cn.ibm.com',
    license="Apache License Version 2.0",
    description="Kubeflow Fairing Python SDK.",
    long_description="Python SDK for Kubeflow Fairing components.",
    url="https://github.com/kubeflow/fairing",
    packages=setuptools.find_packages(
        include=("kubeflow*", "containerregistry*",)),
    package_data={},
    include_package_data=False,
    zip_safe=False,
    classifiers=(
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ),
    install_requires=REQUIRES,
    extras_require={
        'dev': [
            'pytest',
            'pytest-pep8',
            'pytest-cov'
        ]
    }
)
